"""
Attribute inference attacks.
"""

from __future__ import annotations

import json
import logging

# import pickle
from typing import Any

import matplotlib.pyplot as plt
import multiprocess as mp
import numpy as np
from fpdf import FPDF
from sklearn.base import BaseEstimator
from sklearn.preprocessing import OneHotEncoder

from aisdc.attacks import report
from aisdc.attacks.attack import Attack
from aisdc.attacks.dataset import Data

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("aia")

COLOR_A: str = "#86bf91"  # training set plot colour
COLOR_B: str = "steelblue"  # testing set plot colour


class AttributeAttackArgs:
    """Arguments for attribute inference."""

    def __init__(self, **kwargs):
        self.__dict__["report_name"] = None
        self.__dict__["n_cpu"] = max(1, mp.cpu_count() - 1)
        self.__dict__.update(kwargs)

    def __str__(self):
        return ",".join(
            [f"{str(key)}: {str(value)}" for key, value in self.__dict__.items()]
        )

    def set_param(self, key: str, value: Any) -> None:
        """Set a parameter"""
        self.__dict__[key] = value

    def get_args(self) -> dict:
        """Return arguments"""
        return self.__dict__


class AttributeAttack(Attack):
    """Class to wrap the attribute inference attack code."""

    def __init__(self, args: AttributeAttackArgs = AttributeAttackArgs()):
        self.attack_metrics: dict = {}
        self.metadata: dict = {}
        self.args = args

    def __str__(self):
        return "Attribute inference attack"

    def attack(self, dataset: Data, target_model: BaseEstimator) -> None:
        """Programmatic attack entry point

        To be used when code has access to data class and trained target model

        Parameters
        ----------
        dataset: attacks.dataset.Data
            dataset as a Data class object
        target_model: sklearn.base.BaseEstimator
            target model that inherits from an sklearn BaseEstimator
        """
        self.attack_metrics = _attribute_inference(
            target_model, dataset, self.args.n_cpu
        )

    def _construct_metadata(self) -> None:
        """Constructs the metadata object. Called by the reporting method."""
        self.metadata = {}
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"].update(self.args.__dict__)
        self.metadata["attack"] = str(self)

    def make_report(self) -> dict:
        """Create the report.

        Creates the output report. If self.args.report_name is not None, it will also save the
        information in json and pdf formats.

        Returns
        -------

        output: dict
            Dictionary containing all attack output.
        """
        output = {}
        logger.info("Starting report, report_name = %s", self.args.report_name)
        output["attack_metrics"] = self.attack_metrics
        self._construct_metadata()
        output["metadata"] = self.metadata
        if self.args.report_name is not None:
            json_report = json.dumps(output, cls=report.NumpyArrayEncoder)
            with open(f"{self.args.report_name}.json", "w", encoding="utf-8") as f:
                f.write(json_report)
            logger.info("Wrote report to %s", f"{self.args.report_name}.json")

            pdf = create_aia_report(output)
            pdf.output(f"{self.args.report_name}.pdf", "F")
            logger.info("Wrote pdf report to %s", f"{self.args.report_name}.pdf")
        return output


def _unique_max(confidences: list[float], threshold: float) -> bool:
    """Returns whether there is a unique maximum confidence value above
    threshold."""
    if len(confidences) > 0:
        max_conf = np.max(confidences)
        if max_conf < threshold:
            return False
        unique, count = np.unique(confidences, return_counts=True)
        for u, c in zip(unique, count):
            if c == 1 and u == max_conf:
                return True
    return False


def _get_inference_data(  # pylint: disable=too-many-locals
    target_model: BaseEstimator, dataset: Data, feature_id: int, memberset: bool
) -> tuple[np.ndarray, np.ndarray, float]:
    """Returns a dataset of each sample with the attributes to test."""
    attack_feature: dict = dataset.features[feature_id]
    indices: list[int] = attack_feature["indices"]
    unique = np.unique(dataset.x_orig[:, feature_id])
    n_unique: int = len(unique)
    if attack_feature["encoding"] == "onehot":
        onehot_enc = OneHotEncoder()
        values = onehot_enc.fit_transform(unique.reshape(-1, 1)).toarray()
    else:  # pragma: no cover
        # catch all, but can't be reached because this func only called via _infer
        # which is only called for categorical data
        values = unique
    # samples after encoding (e.g. one-hot)
    samples: np.ndarray = dataset.x_train
    # samples before encoding (e.g. str)
    samples_orig: np.ndarray = dataset.x_train_orig
    if not memberset:
        samples = dataset.x_test
        samples_orig = dataset.x_test_orig
    n_samples, x_dim = np.shape(samples)
    x_values = np.zeros((n_samples, n_unique, x_dim), dtype=np.float64)
    y_values = target_model.predict(samples)
    # for each sample to perform inference on
    # add each possible missing feature value
    for i, x in enumerate(samples):
        for j, value in enumerate(values):
            x_values[i][j] = np.copy(x)
            x_values[i][j][indices] = value
    _, counts = np.unique(samples_orig[:, feature_id], return_counts=True)
    baseline = (np.max(counts) / n_samples) * 100
    logger.debug("feature: %d x_values shape = %s", feature_id, np.shape(x_values))
    logger.debug("feature: %d y_values shape = %s", feature_id, np.shape(y_values))
    return x_values, y_values, baseline


def _infer(  # pylint: disable=too-many-locals
    target_model: BaseEstimator,
    dataset: Data,
    feature_id: int,
    threshold: float,
    memberset: bool,
) -> tuple[int, int, float, int, int]:
    """
    For each possible missing value, compute the confidence scores and
    label with the target model; if the label matches the known target model
    label for the original sample, and the highest confidence score is unique,
    infer that attribute if the confidence score is greater than a threshold.
    """
    logger.debug("Commencing attack on feature %d set %d", feature_id, int(memberset))
    correct: int = 0  # number of correct inferences made
    total: int = 0  # total number of inferences made
    x_values, y_values, baseline = _get_inference_data(
        target_model, dataset, feature_id, memberset
    )
    n_unique: int = len(x_values[1])
    samples = dataset.x_train if memberset else dataset.x_test
    for i, x in enumerate(x_values):  # each sample to perform inference on
        # get model confidence scores for all possible values for the sample
        confidence = target_model.predict_proba(x)
        conf = []  # confidences for each possible value with correct label
        attr = []  # features for each possible value with correct label
        # for each possible attribute value,
        # if the label matches the known target model label
        # then store the confidence score and the tested feature vector
        for j in range(n_unique):
            this_label = np.argmax(confidence[j])
            scores = confidence[j][this_label]
            if this_label == y_values[i]:
                conf.append(scores)
                attr.append(x[j])
        # is there is a unique maximum confidence score above threshold?
        if _unique_max(conf, threshold):
            total += 1
            if (attr[np.argmax(conf)] == samples[i]).all():
                correct += 1
    logger.debug("Finished attacking feature %d", feature_id)
    return correct, total, baseline, n_unique, len(samples)


def report_categorical(results: dict) -> str:
    """Returns a string report of the categorical results."""
    results = results["categorical"]
    msg = ""
    for feature in results:
        name = feature["name"]
        _, _, _, n_unique, _ = feature["train"]
        msg += f"Attacking categorical feature {name} with {n_unique} unique values:\n"
        for tranche in ("train", "test"):
            correct, total, baseline, _, n_samples = feature[tranche]
            if total > 0:
                msg += (
                    f"Correctly inferred {(correct / total) * 100:.2f}% "
                    f"of {(total / n_samples) * 100:.2f}% of the {tranche} set; "
                    f"baseline: {baseline:.2f}%\n"
                )
            else:  # pragma: no cover
                # no examples with test dataset where this doesn't happen
                msg += f"Unable to make any inferences of the {tranche} set\n"
    return msg


def report_quantitative(results: dict) -> str:
    """Returns a string report of the quantitative results."""
    results = results["quantitative"]
    msg = ""
    for feature in results:
        msg += (
            f"{feature['name']}: "
            f"{feature['train']:.2f} train risk, "
            f"{feature['test']:.2f} test risk\n"
        )
    return msg


def plot_quantitative_risk(res: dict, savefile: str = "") -> None:
    """Generates bar chart showing quantitative value risk scores."""
    logger.debug("Plotting quantitative feature risk scores")
    results = res["quantitative"]
    if len(results) < 1:  # pragma: no cover
        return
    x = np.arange(len(results))
    ya = []
    yb = []
    names = []
    for feature in results:
        names.append(feature["name"])
        ya.append(feature["train"] * 100)
        yb.append(feature["test"] * 100)
    fig, ax = plt.subplots(1, 1, figsize=(8, 5))
    ax.set_xticks(x)
    ax.set_xticklabels(names, rotation=90)
    ax.set_ylim([0, 100])
    ax.bar(x + 0.2, ya, 0.4, align="center", color=COLOR_A, label="train set")
    ax.bar(x - 0.2, yb, 0.4, align="center", color=COLOR_B, label="test set")
    title = "Percentage of Set at Risk for Quantitative Attributes"
    ax.set_title(f"{res['name']}\n{title}")
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)
    ax.grid(linestyle="dotted", linewidth=1)
    ax.legend(loc="best")
    plt.margins(y=0)
    plt.tight_layout()
    if savefile != "":
        fig.savefig(savefile + "_quant_risk.png", pad_inches=0, bbox_inches="tight")
        logger.debug("Saved quantitative risk plot: %s", savefile)
    else:  # pragma: no cover
        plt.show()


def plot_categorical_risk(  # pylint: disable=too-many-locals
    res: dict, savefile: str = ""
) -> None:
    """Generates bar chart showing categorical risk scores."""
    logger.debug("Plotting categorical feature risk scores")
    results: list[dict] = res["categorical"]
    if len(results) < 1:  # pragma: no cover
        return
    x: np.ndarray = np.arange(len(results))
    ya: list[float] = []
    yb: list[float] = []
    names: list[str] = []
    for feature in results:
        names.append(feature["name"])
        correct_a, total_a, baseline_a, _, _ = feature["train"]
        correct_b, total_b, baseline_b, _, _ = feature["test"]
        a = ((correct_a / total_a) * 100) - baseline_a if total_a > 0 else 0
        b = ((correct_b / total_b) * 100) - baseline_b if total_b > 0 else 0
        ya.append(a)
        yb.append(b)
    fig, ax = plt.subplots(1, 1, figsize=(8, 5))
    ax.set_xticks(x)
    ax.set_xticklabels(names, rotation=90)
    ax.set_ylim([-100, 100])
    ax.bar(x + 0.2, ya, 0.4, align="center", color=COLOR_A, label="train set")
    ax.bar(x - 0.2, yb, 0.4, align="center", color=COLOR_B, label="test set")
    title: str = "Improvement Over Most Common Value Estimate"
    ax.set_title(f"{res['name']}\n{title}")
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)
    ax.grid(linestyle="dotted", linewidth=1)
    ax.legend(loc="best")
    plt.margins(y=0)
    plt.tight_layout()
    if savefile != "":
        fig.savefig(savefile + "_cat_risk.png", pad_inches=0, bbox_inches="tight")
        logger.debug("Saved categorical risk plot: %s", savefile)
    else:  # pragma: no cover
        plt.show()


def plot_categorical_fraction(  # pylint: disable=too-many-locals
    res: dict, savefile: str = ""
) -> None:
    """Generates bar chart showing fraction of dataset inferred."""
    logger.debug("Plotting categorical feature tranche sizes")
    results: list[dict] = res["categorical"]
    if len(results) < 1:  # pragma: no cover
        return
    x: np.ndarray = np.arange(len(results))
    ya: list[float] = []
    yb: list[float] = []
    names: list[str] = []
    for feature in results:
        names.append(feature["name"])
        _, total_a, _, _, n_samples_a = feature["train"]
        _, total_b, _, _, n_samples_b = feature["test"]
        a = ((total_a / n_samples_a) * 100) if n_samples_a > 0 else 0
        b = ((total_b / n_samples_b) * 100) if n_samples_b > 0 else 0
        ya.append(a)
        yb.append(b)
    fig, ax = plt.subplots(1, 1, figsize=(8, 5))
    ax.set_xticks(x)
    ax.set_xticklabels(names, rotation=90)
    ax.set_ylim([0, 100])
    ax.bar(x + 0.2, ya, 0.4, align="center", color=COLOR_A, label="train set")
    ax.bar(x - 0.2, yb, 0.4, align="center", color=COLOR_B, label="test set")
    title: str = "Percentage of Set at Risk"
    ax.set_title(f"{res['name']}\n{title}")
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)
    ax.grid(linestyle="dotted", linewidth=1)
    ax.legend(loc="best")
    plt.margins(y=0)
    plt.tight_layout()
    if savefile != "":
        fig.savefig(savefile + "_cat_frac.png", pad_inches=0, bbox_inches="tight")
        logger.debug("Saved categorical fraction plot: %s", savefile)
    else:  # pragma: no cover
        plt.show()


# def plot_from_file(filename: str, savefile: str = "") -> None: #pragma: no cover
#     """Loads a results save file and plots risk scores.
#        Has been tested but not iuncluded in unit tests or coverage
#        at this stage
#     """
#     logger.debug("Loading from results file: %s", filename)
#     with open(filename + ".pickle", "rb") as handle:
#         results = pickle.load(handle)
#     plot_categorical_risk(results, savefile=savefile)
#     plot_categorical_fraction(results, savefile=savefile)
#     plot_quantitative_risk(results, savefile=savefile)


def _infer_categorical(
    target_model: BaseEstimator, dataset: Data, feature_id: int, threshold: float
) -> dict:
    """Returns the training and test set risks of a categorical feature."""
    result: dict = {
        "name": dataset.features[feature_id]["name"],
        "train": _infer(target_model, dataset, feature_id, threshold, True),
        "test": _infer(target_model, dataset, feature_id, threshold, False),
    }
    return result


def _is_categorical(dataset: Data, feature_id: int) -> bool:
    """Returns whether a feature is categorical.
    For simplicity, assumes integer datatypes are categorical."""
    encoding: str = dataset.features[feature_id]["encoding"]
    if encoding[:3] in ("str", "int") or encoding[:6] in ("onehot"):
        return True
    return False


def _attack_brute_force(
    target_model: BaseEstimator,
    dataset: Data,
    features: list[int],
    n_cpu: int,
    attack_threshold: float = 0,
) -> list[dict]:
    """
    Performs a brute force attribute inference attack by computing the target
    model confidence scores for every value in the list and making an inference
    if there is a unique highest confidence score that exceeds attack_threshold.
    """
    logger.debug("Brute force attacking categorical features")
    args = [
        (target_model, dataset, feature_id, attack_threshold) for feature_id in features
    ]
    with mp.Pool(processes=n_cpu) as pool:  # pylint:disable=not-callable
        results = pool.starmap(_infer_categorical, args)
    return results


def _get_bounds_risk_for_sample(  # pylint: disable=too-many-locals,too-many-arguments
    target_model: BaseEstimator,
    feat_id: int,
    feat_min: float,
    feat_max: float,
    sample: np.ndarray,
    c_min: float = 0,
    protection_limit: float = 0.1,
    feat_n: int = 100,
) -> bool:
    """Returns a bool based on conditions surrounding upper and lower bounds of
    guesses that would lead to the same model confidence.

    Parameters
    ----------

    target_model: BaseEstimator
        Trained target model.
    feat_id: int
        Index of missing feature.
    feat_min: float
        Minimum value of missing feature.
    feat_max: float
        Maximum value of missing feature.
    sample: np.ndarray
        Original known feature vector.
    c_min: float
        Defines the confidence threshold below which we say we don't care.
    protection_limit: float
        Lower [upper] bound on estimated value must not be
        above[below] lower[upper] bounds e.g. 10% of value.
    feat_n: int
        Number of attribute values to test per sample.

    Returns
    -------
    A bool representing whether the quantitative feature is at risk for the sample.

    """
    # attribute values to test - linearly sampled
    x_feat = np.linspace(feat_min, feat_max, feat_n, endpoint=True)
    # get known label
    label: int = int(target_model.predict(sample.reshape(1, -1))[0])
    # a matrix containing feature vector with linearly spaced target attribute
    x1 = np.repeat(sample.reshape(1, -1), feat_n, axis=0)
    x1[:, feat_id] = x_feat
    # get the target model confidences across the attribute range
    confidences = target_model.predict_proba(x1)
    scores = confidences[:, label]  # scores just for the model predicted label
    peak: float = np.max(scores)
    # find lowest and highest values with peak confidence
    lower_bound_index: int = 0
    while scores[lower_bound_index] < peak:
        lower_bound_index += 1
    upper_bound_index: int = feat_n - 1
    while scores[upper_bound_index] < peak:
        upper_bound_index -= 1
    # condition 1: confidence in prediction above some threshold
    # condition 2: confidence for true value == max_confidence
    # condition 3: lower bound above lower protection limit
    # condition 4: upper bound of estimate below upper protection limit
    actual_value = sample[feat_id]
    actual_probs = target_model.predict_proba(sample.reshape(1, -1))[0]
    lower_bound: float = x_feat[lower_bound_index]
    upper_bound: float = x_feat[upper_bound_index]
    if (
        peak > c_min
        and actual_probs[label] == peak
        and lower_bound >= (1 - protection_limit) * actual_value
        and upper_bound <= (1 + protection_limit) * actual_value
    ):  # pragma: no cover
        # has been tested and shown to work in examples notebooks
        # but tests use simplistic nursery dataset
        return True
    return False


def _get_bounds_risk_for_feature(
    target_model: BaseEstimator, feature_id: int, samples: np.ndarray
) -> float:
    """Returns the average feature risk score over a set of samples."""
    feature_risk: int = 0
    n_samples: int = len(samples)
    feat_min: float = np.min(samples[:, feature_id])
    feat_max: float = np.max(samples[:, feature_id])
    for i in range(n_samples):
        sample = samples[i]
        risk = _get_bounds_risk_for_sample(
            target_model, feature_id, feat_min, feat_max, sample
        )
        if risk:  # pragma:no cover
            # can be seen working in examples
            # testing uses nursery with dummy cont. feature
            # which is not predictive
            feature_risk += 1
    if n_samples < 1:  # pragma: no cover
        # is unreachable because of how it is called
        return 0
    return feature_risk / n_samples


def _get_bounds_risk(
    target_model: BaseEstimator,
    feature_name: str,
    feature_id: int,
    x_train: np.ndarray,
    x_test: np.ndarray,
) -> dict:
    """Returns a dictionary containing the training and test set risks of a
    quantitative feature."""
    risk: dict = {
        "name": feature_name,
        "train": _get_bounds_risk_for_feature(target_model, feature_id, x_train),
        "test": _get_bounds_risk_for_feature(target_model, feature_id, x_test),
    }
    return risk


def _get_bounds_risks(
    target_model: BaseEstimator, dataset: Data, features: list[int], n_cpu: int
) -> list[dict]:
    """Computes the bounds risk for all specified features."""
    logger.debug("Computing bounds risk for all specified features")
    args = [
        (
            target_model,
            dataset.features[feature_id]["name"],
            feature_id,
            dataset.x_train,
            dataset.x_test,
        )
        for feature_id in features
    ]
    with mp.Pool(processes=n_cpu) as pool:  # pylint:disable=not-callable
        results = pool.starmap(_get_bounds_risk, args)
    return results


def _attribute_inference(
    target_model: BaseEstimator,
    dataset: Data,
    n_cpu: int,
) -> dict:
    """
    Execute attribute inference attacks on a dataset given a trained model.
    """
    # brute force attack categorical attributes using dataset unique values
    logger.debug("Attacking dataset: %s", dataset.name)
    logger.debug("Attacking categorical attributes...")
    feature_list: list[int] = []
    for feature in range(dataset.n_features):
        if _is_categorical(dataset, feature):
            feature_list.append(feature)
    results_a: list[dict] = _attack_brute_force(
        target_model, dataset, feature_list, n_cpu
    )
    # compute risk scores for quantitative attributes
    logger.debug("Attacking quantitative attributes...")
    feature_list = []
    for feature in range(dataset.n_features):
        if not _is_categorical(dataset, feature):
            feature_list.append(feature)
    results_b: list[dict] = _get_bounds_risks(
        target_model, dataset, feature_list, n_cpu
    )
    # combine results into single object
    results: dict = {
        "name": dataset.name,
        "categorical": results_a,
        "quantitative": results_b,
    }
    return results


def create_aia_report(output: dict, name: str = "aia_report") -> FPDF:
    """Creates PDF report."""
    aia_metrics = output["attack_metrics"]
    metadata = output["metadata"]
    plot_categorical_risk(aia_metrics, name)
    plot_categorical_fraction(aia_metrics, name)
    plot_quantitative_risk(aia_metrics, name)
    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    report.title(pdf, "Attribute Inference Attack Report")
    report.subtitle(pdf, "Introduction")
    report.subtitle(pdf, "Metadata")
    for key, value in metadata["experiment_details"].items():
        report.line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    report.subtitle(pdf, "Metrics")
    categ_rep = report_categorical(aia_metrics).split("\n")
    quant_rep = report_quantitative(aia_metrics).split("\n")
    report.line(pdf, "Categorical Features:", font="courier")
    for line in categ_rep:
        report.line(pdf, line, font="courier")
    report.line(pdf, "Quantitative Features:", font="courier")
    for line in quant_rep:
        report.line(pdf, line, font="courier")
    pdf.add_page()
    report.subtitle(pdf, "Plots")
    if len(aia_metrics["categorical"]) > 0:
        pdf.image(name + "_cat_risk.png", x=None, y=None, w=150, h=0, type="", link="")
        pdf.image(name + "_cat_frac.png", x=None, y=None, w=150, h=0, type="", link="")
    if len(aia_metrics["quantitative"]) > 0:
        pdf.image(
            name + "_quant_risk.png", x=None, y=None, w=150, h=0, type="", link=""
        )
    return pdf



"""
Likelihood testing scenario from https://arxiv.org/pdf/2112.03570.pdf
"""
# pylint: disable = invalid-name
# pylint: disable = too-many-branches

from __future__ import annotations

import argparse
import importlib
import json
import logging
from collections.abc import Hashable, Iterable
from typing import Any

import numpy as np
import sklearn
from scipy.stats import norm
from sklearn.datasets import load_breast_cancer
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from aisdc.attacks import metrics, report
from aisdc.attacks.attack import Attack
from aisdc.attacks.dataset import Data

logging.basicConfig(level=logging.INFO)

N_SHADOW_MODELS = 100  # Number of shadow models that should be trained
EPS = 1e-16  # Used to avoid numerical issues in logit function
P_THRESH = 0.05  # default significance threshold


class DummyClassifier:
    """A Dummy Classifier to allow this code to work with get_metrics"""

    def predict(self, test_X):
        """Return an array of 1/0 depending on value in second column"""
        return 1 * (test_X[:, 1] > 0.5)

    def predict_proba(self, test_X):
        """Simply return the test_X"""
        return test_X


def _logit(p: float) -> float:
    """Standard logit function

    Parameters
    ----------
    p: float
        value to evaluate logit at

    Returns
    -------
    li: float
        logit(p)

    Notes
    -----
    If p is close to 0 or 1, evaluating the log will result in numerical instabilities.
    This code thresholds p at EPS and 1 - EPS where EPS defaults at 1e-16.
    """
    if p > 1 - EPS:
        p = 1 - EPS
    p = max(p, EPS)
    li = np.log(p / (1 - p))
    return li


class LIRAAttackArgs:
    """LIRA Attack arguments"""

    def __init__(self, **kwargs):
        self.__dict__["n_shadow_models"] = N_SHADOW_MODELS
        self.__dict__["p_thresh"] = 0.05
        self.__dict__["report_name"] = None
        self.__dict__["json_file"] = "config.json"
        self.__dict__.update(kwargs)

    def __str__(self):
        return ",".join(
            [f"{str(key)}: {str(value)}" for key, value in self.__dict__.items()]
        )

    def set_param(self, key: Hashable, value: Any) -> None:
        """Set a parameter"""
        self.__dict__[key] = value

    def get_args(self) -> dict:
        """Return arguments"""
        return self.__dict__


class LIRAAttack(Attack):
    """The main LIRA Attack class"""

    def __init__(self, args: LIRAAttackArgs = LIRAAttackArgs()) -> None:
        self.attack_metrics = None
        self.dummy_attack_metrics = None
        self.metadata = None
        self.args = args

    def __str__(self):
        return "LIRA Attack"

    def attack(self, dataset: Data, target_model: sklearn.base.BaseEstimator) -> None:
        """Programmatic attack running
        Runs a LIRA attack from a dataset object and a target model

        Parameters
        ----------
        dataset: attacks.dataset.Data
            Dataset as an instance of the Data class. Needs to have x_train, x_test, y_train
            and y_test set.
        target_mode: sklearn.base.BaseEstimator
            Trained target model. Any class that implements the sklearn.base.BaseEstimator
            interface (i.e. has fit, predict and predict_proba methods)
        """

        shadow_clf = sklearn.base.clone(target_model)

        dataset = self._check_and_update_dataset(dataset, target_model)

        self.run_scenario_from_preds(
            shadow_clf,
            dataset.x_train,
            dataset.y_train,
            target_model.predict_proba(dataset.x_train),
            dataset.x_test,
            dataset.y_test,
            target_model.predict_proba(dataset.x_test),
        )

    def _check_and_update_dataset(
        self, dataset: Data, target_model: sklearn.base.BaseEstimator
    ) -> tuple[np.ndarray, np.ndarray]:
        """Makes sure that it is ok to use the class variables to index the prediction
        arrays. This has two steps:
        1. Replacing the values in y_train with their position in target_model.classes (will
           normally result in no change)
        2. Removing from the test set any rows corresponding to classes that are not in the
           training set.
        """

        y_train_new = []
        classes = list(target_model.classes_)  # pylint: disable = protected-access
        for y in dataset.y_train:
            y_train_new.append(classes.index(y))

        dataset.y_train = np.array(y_train_new, int)
        print(
            f"new ytrain has values and counts: {np.unique(dataset.y_train,return_counts=True)}"
        )
        ok_pos = []
        y_test_new = []
        for i, y in enumerate(dataset.y_test):
            if y in classes:
                ok_pos.append(i)
                y_test_new.append(classes.index(y))

        if len(y_test_new) != len(dataset.x_test):
            dataset.x_test = dataset.x_test[ok_pos, :]
        dataset.y_test = np.array(y_test_new, int)
        print(
            f"new ytest has values and counts: {np.unique(dataset.y_test,return_counts=True)}"
        )

        return dataset

    def run_scenario_from_preds(  # pylint: disable = too-many-statements
        self,
        shadow_clf: sklearn.base.BaseEstimator,
        X_target_train: Iterable[float],
        y_target_train: Iterable[float],
        target_train_preds: Iterable[float],
        X_shadow_train: Iterable[float],
        y_shadow_train: Iterable[float],
        shadow_train_preds: Iterable[float],
    ) -> tuple[np.ndarray, np.ndarray, sklearn.base.BaseEstimator]:
        # def run_scenario_from_preds( # pylint: disable = too-many-locals, too-many-arguments
        #     shadow_clf,
        #     X_target_train: Iterable[float],
        #     y_target_train: Iterable[float],
        #     target_train_preds: Iterable[float],
        #     X_shadow_train: Iterable[float],
        #     y_shadow_train: Iterable[float],
        #     shadow_train_preds: Iterable[float],

        """Implements the likelihood test, using the "offline" version
        See p.6 (top of second column) for details

        Parameters
        ----------
        shadow_clf: sklearn.Model
            An sklearn classifier that will be trained to form the shadow model.
            All hyper-parameters should have been set.
        X_target_train: np.ndarray
            Data that was used to train the target model
        y_target_train: np.ndarray
            Labels that were used to train the target model
        target_train_preds: np.ndarray
            Array of predictions produced by the target model on the training data
        X_shadow_train: np.ndarray
            Data that will be used to train the shadow models
        y_shadow_train: np.ndarray
            Labels that will be used to train the shadow model
        shadow_train_preds: np.ndarray
            Array of predictions produced by the target model on the shadow data


        Returns
        -------
        mia_scores: np.ndarray
            Attack probabilities of belonging to the training set or not
        mia_labels: np.ndarray
            True labels of belonging to the training set or not
        mia_cls: DummyClassifier
            A DummyClassifier that directly returns the scores for compatibility with code
            in metrics.py

        Examples
        --------

        >>> X, y = load_breast_cancer(return_X_y=True, as_frame=False)
        >>> train_X, test_X, train_y, test_y = train_test_split(
        >>>   X, y, test_size=0.5, stratify=y
        >>> )
        >>> rf = RandomForestClassifier(min_samples_leaf=1, min_samples_split=2)
        >>> rf.fit(train_X, train_y)
        >>> mia_test_probs, mia_test_labels, mia_clf = likelihood_scenario(
        >>>     RandomForestClassifier(min_samples_leaf=1, min_samples_split=2, max_depth=10),
        >>>     train_X,
        >>>     train_y,
        >>>     rf.predict_proba(train_X),
        >>>     test_X,
        >>>     test_y,
        >>>     rf.predict_proba(test_X),
        >>>     n_shadow_models=100
        >>> )
        """

        logger = logging.getLogger("lr-scenario")

        n_train_rows, _ = X_target_train.shape
        n_shadow_rows, _ = X_shadow_train.shape
        indices = np.arange(0, n_train_rows + n_shadow_rows, 1)

        # Combine taregt and shadow train, from which to sample datasets
        combined_X_train = np.vstack((X_target_train, X_shadow_train))
        combined_y_train = np.hstack((y_target_train, y_shadow_train))

        train_row_to_confidence = {i: [] for i in range(n_train_rows)}
        shadow_row_to_confidence = {i: [] for i in range(n_shadow_rows)}

        # Train N_SHADOW_MODELS shadow models
        logger.info("Training shadow models")
        for model_idx in range(self.args.n_shadow_models):
            if model_idx % 10 == 0:
                logger.info("Trained %d models", model_idx)
            # Pick the indices to use for training this one
            np.random.seed(model_idx)  # Reproducibility
            these_idx = np.random.choice(indices, n_train_rows, replace=False)
            temp_X_train = combined_X_train[these_idx, :]
            temp_y_train = combined_y_train[these_idx]

            # Fit the shadow model
            shadow_clf.set_params(random_state=model_idx)
            shadow_clf.fit(temp_X_train, temp_y_train)

            # map a class to a column
            class_map = {c: i for i, c in enumerate(shadow_clf.classes_)}

            # Get the predicted probabilities on the training data
            confidences = shadow_clf.predict_proba(X_target_train)
            # print(f'shadow clf returned confidences with shape {confidences.shape}')

            these_idx = set(these_idx)
            for i in range(n_train_rows):
                if i not in these_idx:
                    # If i was _not_ used for training, incorporate the logit of its confidence of
                    # being correct - TODO: should we just be taking max??
                    cl_pos = class_map.get(y_target_train[i], -1)
                    # Occasionally, the random data split will result in classes being
                    # absent from the training set. In these cases cl_pos will be -1 and
                    # we include logit(0) instead of discarding (also for the shadow data below)
                    if cl_pos >= 0:
                        train_row_to_confidence[i].append(
                            _logit(confidences[i, cl_pos])
                        )
                    else:  # pragma: no cover
                        # catch-all
                        train_row_to_confidence[i].append(_logit(0))
            # Same process for shadow data
            shadow_confidences = shadow_clf.predict_proba(X_shadow_train)
            for i in range(n_shadow_rows):
                if i + n_train_rows not in these_idx:
                    cl_pos = class_map.get(y_shadow_train[i], -1)
                    if cl_pos >= 0:
                        shadow_row_to_confidence[i].append(
                            _logit(shadow_confidences[i, cl_pos])
                        )
                    else:  # pragma: no cover
                        # catch-all
                        shadow_row_to_confidence[i].append(_logit(0))

        # Do the test described in the paper in each case
        mia_scores = []
        mia_labels = []
        logger.info("Computing scores for train rows")
        for i in range(n_train_rows):
            true_score = _logit(target_train_preds[i, y_target_train[i]])
            null_scores = np.array(train_row_to_confidence[i])
            mean_null = 0.0
            var_null = 0.0
            if not np.isnan(null_scores).all():
                mean_null = np.nanmean(null_scores)  # null_scores.mean()
                var_null = np.nanvar(null_scores)  #
            var_null = max(var_null, EPS)  # var can be zero in some cases
            prob = norm.cdf(true_score, loc=mean_null, scale=np.sqrt(var_null))
            mia_scores.append([1 - prob, prob])
            mia_labels.append(1)

        logger.info("Computing scores for shadow rows")
        for i in range(n_shadow_rows):
            true_score = _logit(shadow_train_preds[i, y_shadow_train[i]])
            null_scores = np.array(shadow_row_to_confidence[i])
            mean_null = null_scores.mean()
            var_null = max(null_scores.var(), EPS)  # var can be zeros in some cases
            prob = norm.cdf(true_score, loc=mean_null, scale=np.sqrt(var_null))
            mia_scores.append([1 - prob, prob])
            mia_labels.append(0)

        mia_clf = DummyClassifier()
        logger.info("Finished scenario")
        self.attack_metrics = [
            metrics.get_metrics(mia_clf, np.array(mia_scores), np.array(mia_labels))
        ]

    def example(self) -> None:  # pylint: disable = too-many-locals
        """Runs an example attack using data from sklearn

        Generates example data, trains a classifier and tuns the attack
        """
        X, y = load_breast_cancer(return_X_y=True, as_frame=False)
        train_X, test_X, train_y, test_y = train_test_split(
            X, y, test_size=0.5, stratify=y
        )
        rf = RandomForestClassifier(min_samples_leaf=1, min_samples_split=2)
        rf.fit(train_X, train_y)
        self.run_scenario_from_preds(
            sklearn.base.clone(rf),
            train_X,
            train_y,
            rf.predict_proba(train_X),
            test_X,
            test_y,
            rf.predict_proba(test_X),
        )

    def _construct_metadata(self) -> None:
        """Constructs the metadata object. Called by the reporting method"""
        self.metadata = {}
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"].update(self.args.__dict__)
        if "func" in self.metadata["experiment_details"]:
            del self.metadata["experiment_details"]["func"]

        self.metadata["global_metrics"] = {}

        pdif = np.exp(-self.attack_metrics[0]["PDIF01"])

        self.metadata["global_metrics"]["PDIF_sig"] = (
            f"Significant at p={self.args.p_thresh}"
            if pdif <= self.args.p_thresh
            else f"Not significant at p={self.args.p_thresh}"
        )

        auc_p, auc_std = metrics.auc_p_val(
            self.attack_metrics[0]["AUC"],
            self.attack_metrics[0]["n_pos_test_examples"],
            self.attack_metrics[0]["n_neg_test_examples"],
        )
        self.metadata["global_metrics"]["AUC_sig"] = (
            f"Significant at p={self.args.p_thresh}"
            if auc_p <= self.args.p_thresh
            else f"Not significant at p={self.args.p_thresh}"
        )
        self.metadata["global_metrics"][
            "AUC NULL 3sd range"
        ] = f"{0.5 - 3 * auc_std} -> {0.5 + 3 * auc_std}"

        self.metadata["attack"] = str(self)

    def make_report(self) -> dict:
        """Create the report

        Creates the output report. If self.args.report_name is not None, it will also save the
        information in json and pdf formats

        Returns
        -------

        output: Dict
            Dictionary containing all attack output
        """
        logger = logging.getLogger("reporting")
        output = {}
        logger.info("Starting report, report_name = %s", self.args.report_name)
        output["attack_metrics"] = self.attack_metrics
        self._construct_metadata()
        output["metadata"] = self.metadata
        if self.args.report_name is not None:
            json_report = report.create_json_report(output)
            with open(f"{self.args.report_name}.json", "w", encoding="utf-8") as f:
                f.write(json_report)
            logger.info("Wrote report to %s", f"{self.args.report_name}.json")

            pdf = report.create_lr_report(output)
            pdf.output(f"{self.args.report_name}.pdf", "F")
            logger.info("Wrote pdf report to %s", f"{self.args.report_name}.pdf")
        return output

    def setup_example_data(self) -> None:
        """Method to create example data and save (including config). Intended to allow users
        to see how they would need to setup their own data.

        Generates train and test data .csv files, train and test predictions .csv files and
        a config.json file that can be used to run the attack from the command line.
        """
        X, y = load_breast_cancer(return_X_y=True)
        train_X, test_X, train_y, test_y = train_test_split(
            X, y, test_size=0.5, stratify=y
        )
        rf = RandomForestClassifier(min_samples_split=2, min_samples_leaf=1)
        rf.fit(train_X, train_y)
        train_data = np.hstack((train_X, train_y[:, None]))
        np.savetxt("train_data.csv", train_data, delimiter=",")

        test_data = np.hstack((test_X, test_y[:, None]))
        np.savetxt("test_data.csv", test_data, delimiter=",")

        train_preds = rf.predict_proba(train_X)
        test_preds = rf.predict_proba(test_X)
        np.savetxt("train_preds.csv", train_preds, delimiter=",")
        np.savetxt("test_preds.csv", test_preds, delimiter=",")

        config = {
            "training_data_file": "train_data.csv",
            "testing_data_file": "test_data.csv",
            "training_preds_file": "train_preds.csv",
            "testing_preds_file": "test_preds.csv",
            "target_model": ["sklearn.ensemble", "RandomForestClassifier"],
            "target_hyppars": {"min_samples_split": 2, "min_samples_leaf": 1},
        }

        with open("config.json", "w", encoding="utf-8") as f:
            f.write(json.dumps(config))

    def attack_from_config(self) -> None:  # pylint: disable = too-many-locals
        """Runs an attack based on the args parsed from the command line"""
        logger = logging.getLogger("run-attack")
        logger.info("Reading config from %s", self.args.json_file)
        with open(self.args.json_file, encoding="utf-8") as f:
            config = json.loads(f.read())

        logger.info("Loading training data csv from %s", config["training_data_file"])
        training_data = np.loadtxt(config["training_data_file"], delimiter=",")
        train_X = training_data[:, :-1]
        train_y = training_data[:, -1].flatten().astype(int)
        logger.info("Loaded %d rows", len(train_X))

        logger.info("Loading test data csv from %s", config["testing_data_file"])
        test_data = np.loadtxt(config["testing_data_file"], delimiter=",")
        test_X = test_data[:, :-1]
        test_y = test_data[:, -1].flatten().astype(int)
        logger.info("Loaded %d rows", len(test_X))

        logger.info("Loading train predictions form %s", config["training_preds_file"])
        train_preds = np.loadtxt(config["training_preds_file"], delimiter=",")
        assert len(train_preds) == len(train_X)

        logger.info("Loading test predictions form %s", config["testing_preds_file"])
        test_preds = np.loadtxt(config["testing_preds_file"], delimiter=",")
        assert len(test_preds) == len(test_X)

        clf_module_name, clf_class_name = config["target_model"]
        module = importlib.import_module(clf_module_name)
        clf_class = getattr(module, clf_class_name)
        clf_params = config["target_hyppars"]
        clf = clf_class(**clf_params)
        logger.info("Created model: %s", str(clf))
        self.run_scenario_from_preds(
            clf, train_X, train_y, train_preds, test_X, test_y, test_preds
        )
        logger.info("Computing metrics")


# Methods invoked by command line script
def _setup_example_data(args):
    """Call the methods to setup some example data"""
    lira_args = LIRAAttackArgs(**args.__dict__)
    attack_obj = LIRAAttack(lira_args)
    attack_obj.setup_example_data()


def _example(args):
    """Call the methods to run an example"""
    lira_args = LIRAAttackArgs(**args.__dict__)
    attack_obj = LIRAAttack(lira_args)
    attack_obj.example()
    attack_obj.make_report()


def _run_attack(args):
    """Run a command line attack based on saved files described in .json file"""
    lira_args = LIRAAttackArgs(**args.__dict__)
    attack_obj = LIRAAttack(lira_args)
    attack_obj.attack_from_config()
    attack_obj.make_report()


def main():
    """Main method to parse args and invoke relevant code"""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-s",
        "--n-shadow-models",
        type=int,
        required=False,
        default=N_SHADOW_MODELS,
        action="store",
        dest="n_shadow_models",
        help=("The number of shadow models to train (default = %(default)d)"),
    )
    parser.add_argument(
        "--report-name",
        type=str,
        action="store",
        dest="report_name",
        required=False,
        default="lr_report",
        help=("Output name for the report. Default = %(default)s"),
    )
    parser.add_argument(
        "-p",
        "--p-thresh",
        type=float,
        action="store",
        dest="p_thresh",
        required=False,
        default=P_THRESH,
        help=("Significance threshold for p-value comparisons. Default = %(default)f"),
    )
    subparsers = parser.add_subparsers()
    example_parser = subparsers.add_parser("run-example", parents=[parser])
    example_parser.set_defaults(func=_example)

    attack_parser = subparsers.add_parser("run-attack", parents=[parser])
    attack_parser.add_argument(
        "-j",
        "--json-file",
        action="store",
        required=True,
        dest="json_file",
        type=str,
        help=(
            "Name of the .json file containing details for the run. Default = %(default)s"
        ),
    )
    attack_parser.set_defaults(func=_run_attack)

    example_data_parser = subparsers.add_parser("setup-example-data")
    example_data_parser.set_defaults(func=_setup_example_data)

    args = parser.parse_args()
    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        print(e)
        print("Invalid command. Try --help to get more details")


if __name__ == "__main__":  # pragma:no cover
    main()



"""
loaders.py
A set of useful handlers to pull in datasets common to the project and perform the appropriate
pre-processing
"""

# pylint: disable=import-error, invalid-name, consider-using-with, too-many-return-statements

import logging
import os
from collections import Counter
from typing import List, Tuple
from zipfile import BadZipFile, ZipFile

import numpy as np
import pandas as pd
import pylab as plt
from sklearn.datasets import fetch_openml, load_iris
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

# Following is to stop pylint always sqwarking at pandas things
# pylint: disable = no-member, unsubscriptable-object


logging.basicConfig(level="DEBUG")

logger = logging.getLogger(__file__)


PROJECT_ROOT_FOLDER = os.path.dirname(os.path.dirname(__file__))
logger.info("ROOT PROJECT FOLDER = %s", PROJECT_ROOT_FOLDER)


class UnknownDataset(Exception):
    """Exception raised if the user passes a name that we don't recognise"""


class DataNotAvailable(Exception):
    """Exception raised if the user asks for a dataset that they do not have the data for. I.e.
    some datasets require a .csv file to have been downloaded.
    """


def get_data_sklearn(  # pylint: disable = too-many-branches
    dataset_name: str, data_folder: str = os.path.join(PROJECT_ROOT_FOLDER, "data")
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Main entry method to return data in format sensible for sklearn. User passes a name and that
    dataset is returned as a tuple of pandas DataFrames (data, labels).

    Parameters
    ----------
    dataset_name: str
        The name of the dataset to load
    data_folder: str
        The name of the local folder in which data is stored.

    Returns
    -------

    X: pd.DataFrame
        The input dataframe -- rows are examples, columns variables
    y: pd.DataFrame
        the target dataframe -- has a single column containing the target values

    Notes
    -----

    The following datasets are available:
    mimic2-iaccd (requires data download)
    in-hospital-mortality (requires data download)
    medical-mnist-ab-v-br-100 (requires data download)
    medical-mnist-ab-v-br-500 (requires data download)
    medical-mnist-all-100 (requires data download)
    indian liver (requires data download)
    texas hospitals 10 (requires data download)
    synth-ae (requires data download)
    synth-ae-small (requires data download)
    nursery (downloads automatically)
    iris (available out of the box via sklearn)

    Datasets can be normalised by adding the following prefixes:
    standard: standardises all columns to have zero mean and unit variance.
    minmax: standardises all columns to have values between 0 and 1.
    round: rounds continues features to have 3dp

    These can be nested.

    Examples
    --------
    >>> X, y = get_data_sklearn("mimic2-iaccd") # pull the mimic2-iaccd data
    >>> X, y = get_data_sklearn("minmax iris") # pull the iris data and round continuous features


    """
    logger.info("DATASET FOLDER = %s", data_folder)

    if dataset_name.startswith("standard"):
        sub_name = dataset_name.split("standard")[1].strip()
        feature_df, target_df = get_data_sklearn(sub_name)
        for column in feature_df.columns:
            col_mean = feature_df[column].mean()
            col_std = np.sqrt(feature_df[column].var())
            feature_df[column] = feature_df[column] - col_mean
            feature_df[column] = feature_df[column] / col_std
        return feature_df, target_df

    if dataset_name.startswith("minmax"):
        sub_name = dataset_name.split("minmax")[1].strip()
        feature_df, target_df = get_data_sklearn(sub_name)
        for column in feature_df.columns:
            col_min = feature_df[column].min()
            col_range = feature_df[column].max() - col_min
            feature_df[column] = feature_df[column] - col_min
            feature_df[column] = feature_df[column] / col_range
        return feature_df, target_df

    if dataset_name.startswith("round"):
        sub_name = dataset_name.split("round")[1].strip()
        logger.debug(sub_name)
        feature_df, target_df = get_data_sklearn(sub_name)
        column_dtype = feature_df.dtypes  # pylint: disable = no-member

        for i, column in enumerate(feature_df.columns):
            if column_dtype[i] == "float64":
                feature_df[column] = feature_df[column].round(decimals=3)
        return feature_df, target_df

    if dataset_name == "mimic2-iaccd":
        return _mimic_iaccd(data_folder)
    if dataset_name == "in-hospital-mortality":
        return _in_hospital_mortality(data_folder)
    if dataset_name == "medical-mnist-ab-v-br-100":
        return _medical_mnist_loader(data_folder, 100, ["AbdomenCT", "BreastMRI"])
    if dataset_name == "medical-mnist-ab-v-br-500":
        return _medical_mnist_loader(data_folder, 500, ["AbdomenCT", "BreastMRI"])
    if dataset_name == "medical-mnist-all-100":
        return _medical_mnist_loader(
            data_folder,
            100,
            ["AbdomenCT", "BreastMRI", "CXR", "ChestCT", "Hand", "HeadCT"],
        )
    if dataset_name == "indian liver":
        return _indian_liver(data_folder)
    if dataset_name == "texas hospitals 10":
        return _texas_hospitals(data_folder)
    if dataset_name == "synth-ae":
        return _synth_ae(data_folder)
    if dataset_name == "synth-ae-small":
        return _synth_ae(data_folder, 200)
    if dataset_name == "nursery":
        return _nursery()
    if dataset_name == "iris":
        return _iris()
    raise UnknownDataset(dataset_name)


def _iris() -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    sklearn iris data - just first two classes
    """
    X, y = load_iris(return_X_y=True, as_frame=True)
    X = X[y < 2]
    y = y[y < 2]
    return X, pd.DataFrame(y)


def _nursery() -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    The sklearn nursery dataset
    """

    data = fetch_openml(data_id=26, as_frame=True)

    target_encoder = LabelEncoder()
    target_vals = target_encoder.fit_transform(data["target"].values)
    target_dataframe = pd.DataFrame({"target": target_vals})

    feature_encoder = OneHotEncoder()
    x_encoded = feature_encoder.fit_transform(data["data"]).toarray()
    feature_dataframe = pd.DataFrame(
        x_encoded, columns=feature_encoder.get_feature_names_out()
    )

    return feature_dataframe, target_dataframe


# Patched to support non-flattened images. Same behaviour as before except if called with
# flatten=False explicitly.
def _images_to_ndarray(
    images_dir: str, number_to_load: int, label: int, flatten: bool = True
) -> Tuple[np.array, np.array]:
    """
    Grab number_to_load images from the images_dir and create a np array and label array
    """
    folder_path = images_dir + os.sep
    images_names = sorted(os.listdir(folder_path))
    images_names = images_names[:number_to_load]
    # fix f or macosx
    if ".DS_Store" in images_names:  # pragma: no cover
        images_names.remove(".DS_Store")

    if flatten:
        np_images = np.array(
            [plt.imread(folder_path + img).flatten() for img in images_names]
        )
    else:
        np_images = np.array([plt.imread(folder_path + img) for img in images_names])
    labels = np.ones((len(np_images), 1), int) * label
    return (np_images, labels)


def _medical_mnist_loader(  # pylint: disable = too-many-locals
    data_folder: str, n_per_class: int, classes: List[str]
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Load Medical MNIST into pandas format
    borrows heavily from: https://www.kaggle.com/harelshattenstein/medical-mnist-knn
    Creates a binary classification
    """

    base_folder = os.path.join(
        data_folder,
        "kaggle-medical-mnist",
        "archive",
    )

    zip_file = os.path.join(data_folder, "kaggle-medical-mnist", "archive.zip")

    logger.debug(base_folder, data_folder)
    if not any([os.path.exists(base_folder), os.path.exists(zip_file)]):
        help_message = f"""
Data file {base_folder} does not exist. Please download fhe file from:
https://www.kaggle.com/andrewmvd/medical-mnist
and place it in the correct folder. It unzips the file first.
        """
        raise DataNotAvailable(help_message)

    if os.path.exists(base_folder):
        pass
    elif os.path.exists(zip_file):
        try:
            with ZipFile(zip_file) as zip_handle:
                zip_handle.extractall(base_folder)
                logger.debug("Extracted all")
        except BadZipFile:  # pragma: no cover
            logger.error("Encountered bad zip file")
            raise

    labels_dict = {
        0: "AbdomenCT",
        1: "BreastMRI",
        2: "CXR",
        3: "ChestCT",
        4: "Hand",
        5: "HeadCT",
    }

    reverse_labels_dict = {v: k for k, v in labels_dict.items()}

    for i, class_name in enumerate(classes):
        label = reverse_labels_dict[class_name]
        x_images, y_images = _images_to_ndarray(
            os.path.join(base_folder, class_name), n_per_class, label
        )

        if i == 0:
            all_x = x_images
            all_y = y_images
        else:
            all_x = np.vstack((all_x, x_images))
            all_y = np.vstack((all_y, y_images))

    return (pd.DataFrame(all_x), pd.DataFrame(all_y))


def _synth_ae(
    data_folder: str, n_rows: int = 5000
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    First norws (default 5000) rows from the Synthetic A&E data from NHS England
    https://data.england.nhs.uk/dataset/a-e-synthetic-data/resource/81b068e5-6501-4840-a880-a8e7aa56890e # pylint: disable=line-too-long
    """

    file_path = os.path.join(
        data_folder, "AE_England_synthetic.csv"  #'A&E Synthetic Data.csv'
    )

    if not os.path.exists(file_path):
        help_message = f"""
Data file {file_path} does not exist. Please download the file from:
https://data.england.nhs.uk/dataset/a-e-synthetic-data/resource/81b068e5-6501-4840-a880-a8e7aa56890e
unzip it (7z) and then copy the .csv file into your data folder.
    """
        raise DataNotAvailable(help_message)

    input_data = pd.read_csv(file_path, nrows=n_rows)
    columns_to_drop = [
        "AE_Arrive_Date",
        "AE_Arrive_HourOfDay",
        "Admission_Method",
        "ICD10_Chapter_Code",
        "Treatment_Function_Code",
        "Length_Of_Stay_Days",
        "ProvID",
    ]
    input_data.drop(columns_to_drop, axis=1, inplace=True)

    # Remove any rows with NAs in the remaining columns
    input_data.dropna(axis=0, inplace=True)

    # One-hot encode some columns
    encode_columns = ["Age_Band", "AE_HRG"]
    encode_data = input_data[encode_columns]
    input_data.drop(encode_columns, axis=1, inplace=True)

    oh = OneHotEncoder()
    oh.fit(encode_data)
    onehot_df = pd.DataFrame(
        oh.transform(encode_data).toarray(),
        columns=oh.get_feature_names_out(),
        index=input_data.index,
    )

    input_data = pd.concat([input_data, onehot_df], axis=1)

    X = input_data.drop(["Admitted_Flag"], axis=1)
    y = input_data[["Admitted_Flag"]]

    return (X, y)


def _indian_liver(data_folder: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Indian Liver Patient Dataset
    https://archive.ics.uci.edu/ml/machine-learning-databases/00225/Indian%20Liver%20Patient%20Dataset%20(ILPD).csv # pylint: disable=line-too-long
    """
    # (https://archive.ics.uci.edu/ml/datasets/ILPD+(Indian+Liver+Patient+Dataset)
    file_path = os.path.join(data_folder, "Indian Liver Patient Dataset (ILPD).csv")
    if not os.path.exists(file_path):
        help_message = f"""
Data file {file_path} does not exist. Please download fhe file from:
https://archive.ics.uci.edu/ml/datasets/ILPD+(Indian+Liver+Patient+Dataset
and place it in the correct folder.
        """
        raise DataNotAvailable(help_message)

    column_names = [
        "age",
        "gender",
        "total Bilirubin",
        "direct Bilirubin",
        "Alkphos",
        "SGPT",
        "SGOT",
        "total proteins",
        "albumin",
        "A/G ratio",
        "class",
    ]

    liver_data = pd.read_csv(file_path, names=column_names, index_col=False)

    liver_data.gender.replace("Male", 0, inplace=True)
    liver_data.gender.replace("Female", 1, inplace=True)

    liver_data.dropna(axis=0, inplace=True)

    liver_labels = liver_data["class"]
    liver_data.drop(["class"], axis=1, inplace=True)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(liver_labels.values)
    liver_labels = pd.DataFrame({"class": encoded_labels})
    return (liver_data, liver_labels)


def _in_hospital_mortality(data_folder: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    In-hospital mortality data from this study:
        https://datadryad.org/stash/dataset/doi:10.5061/dryad.0p2ngf1zd
    """
    # Check the data has been downloaded. If not throw an exception with instructions on how to
    # download, and where to store
    files = ["data01.csv", "doi_10.5061_dryad.0p2ngf1zd__v5.zip"]
    file_path = [os.path.join(data_folder, f) for f in files]

    if not any(  # pylint: disable=use-a-generator
        [os.path.exists(fp) for fp in file_path]
    ):  # pylint: disable=use-a-generator
        help_message = f"""
Data file {file_path[0]} or {file_path[1]} does not exist. Please download the file from:
https://datadryad.org/stash/dataset/doi:10.5061/dryad.0p2ngf1zd
and place it in the correct folder. It works with either the zip file or uncompressed.
        """
        raise DataNotAvailable(help_message)

    if os.path.exists(file_path[1]):
        input_data = pd.read_csv(ZipFile(file_path[1]).open("data01.csv"))
    else:
        input_data = pd.read_csv(file_path[0])
    clean_data = input_data.dropna(axis=0, how="any").drop(columns=["group", "ID"])
    target = "outcome"
    labels = clean_data[target]
    features = clean_data.drop([target], axis=1)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(labels.values)
    labels = pd.DataFrame({"outcome": encoded_labels})

    return (features, labels)


def _mimic_iaccd(data_folder: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Loads the mimic_iaccd data and performs Alba's pre-processing
    """

    # Check the data has been downloaded. If not throw an exception with instructions on how to
    # download, and where to store
    file_path = os.path.join(data_folder, "mimic2-iaccd", "1.0", "full_cohort_data.csv")

    if not os.path.exists(file_path):
        help_message = f"""
        The MIMIC2-iaccd data is not available in {data_folder}.
        The following file should exist: {file_path}.
        Please download from https://physionet.org/content/mimic2-iaccd/1.0/full_cohort_data.csv
        """
        raise DataNotAvailable(help_message)

    # File exists, load and preprocess#
    logger.info("Loading mimic2-iaccd")
    input_data = pd.read_csv(file_path)

    logger.info("Preprocessing")
    # remove columns non-numerical and repetitive or uninformative data for the analysis
    col = [
        "service_unit",
        "day_icu_intime",
        "hosp_exp_flg",
        "icu_exp_flg",
        "day_28_flg",
    ]
    # service_num is the numerical version of service_unit
    # day_icu_intime_num is the numerical version of day_icu_intime
    # the other columns are to do with death and are somewhat repetitive with censor_flg
    input_data.drop(col, axis=1, inplace=True)

    # drop columns with only 1 value
    input_data.drop("sepsis_flg", axis=1, inplace=True)

    # drop NA by row
    input_data.dropna(axis=0, inplace=True)

    # extract target
    target = "censor_flg"
    y = input_data[target]
    X = input_data.drop([target, "mort_day_censored"], axis=1)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(y.values)
    y = pd.DataFrame({"censor_flag": encoded_labels})

    return (X, y)


def _texas_hospitals(
    data_folder: str,
) -> Tuple[pd.DataFrame, pd.DataFrame]:  # pragma: no cover
    # pylint: disable=too-many-statements, too-many-locals
    """
    Texas Hospitals Dataset
    (https://www.dshs.texas.gov/THCIC/Hospitals/Download.shtm)

    Note: this has been tested repeated in the GRAIMatter project.
    However, for licensing reasons we cannot redistribute the data.
    Therefore it is omitted from CI test coverage and metrics.

    """
    file_list = [
        "PUDF-1Q2006-tab-delimited.zip",
        "PUDF-1Q2007-tab-delimited.zip",
        "PUDF-1Q2008-tab-delimited.zip",
        "PUDF-1Q2009-tab-delimited.zip",
        "PUDF-2Q2006-tab-delimited.zip",
        "PUDF-2Q2007-tab-delimited.zip",
        "PUDF-2Q2008-tab-delimited.zip",
        "PUDF-2Q2009-tab-delimited.zip",
        "PUDF-3Q2006-tab-delimited.zip",
        "PUDF-3Q2007-tab-delimited.zip",
        "PUDF-3Q2008-tab-delimited.zip",
        "PUDF-3Q2009-tab-delimited.zip",
        "PUDF-4Q2006-tab-delimited.zip",
        "PUDF-4Q2007-tab-delimited.zip",
        "PUDF-4Q2008-tab-delimited.zip",
        "PUDF-4Q2009-tab-delimited.zip",
    ]

    files_path = [os.path.join(data_folder, "TexasHospitals", f) for f in file_list]

    found = [os.path.exists(file_path) for file_path in files_path]
    not_found = [file_path for file_path in files_path if not os.path.exists(file_path)]

    processed_data_file = "texas_data10_rm_binary.csv"
    if not all(found):
        help_message = f"""
    Some or all data files do not exist. Please accept their terms & conditions,then download the
    tab delimited files from each quarter during 2006-2009 from:
    https://www.dshs.texas.gov/THCIC/Hospitals/Download.shtm
and place it in the correct folder.

    Missing files are:
    {not_found}
        """
        raise DataNotAvailable(help_message)

    if not os.path.exists(
        os.path.join(data_folder, "TexasHospitals", processed_data_file)
    ):
        logger.info("Processing Texas Hospitals data (2006-2009)")

        # Load data
        columns_names = [
            "THCIC_ID",  # Provider ID. Unique identifier assigned to the provider by DSHS.
            # Hospitals with fewer than 50 discharges have been aggregated into the
            # Provider ID '999999'
            "DISCHARGE_QTR",  # yyyyQm
            "TYPE_OF_ADMISSION",
            "SOURCE_OF_ADMISSION",
            "PAT_ZIP",  # Patient’s five-digit ZIP code
            "PUBLIC_HEALTH_REGION",  # Public Health Region of patient’s address
            "PAT_STATUS",  # Code indicating patient status as of the ending date of service for
            # the period of care reported
            "SEX_CODE",
            "RACE",
            "ETHNICITY",
            "LENGTH_OF_STAY",
            "PAT_AGE",  # Code indicating age of patient in days or years on date of discharge.
            "PRINC_DIAG_CODE",  # diagnosis code for the principal diagnosis
            "E_CODE_1",  # external cause of injury
            "PRINC_SURG_PROC_CODE",  # Code for the principal surgical or other procedure performed
            # during the period covered by the bill
            "RISK_MORTALITY",  # Assignment of a risk of mortality score from the All Patient
            # Refined (APR) Diagnosis Related Group (DRG)
            "ILLNESS_SEVERITY",  # Assignment of a severity of illness score from the All Patient
            # Refined (APR) Diagnosis RelatedGroup (DRG
            "RECORD_ID",
        ]
        # obtain the 100 most frequent procedures
        tmp = []
        for f in files_path:
            df = [
                pd.read_csv(
                    ZipFile(f).open(i), sep="\t", usecols=["PRINC_SURG_PROC_CODE"]
                )
                for i in ZipFile(f).namelist()
                if "base" in i
            ][0]
            df.dropna(inplace=True)
            tmp.extend(list(df.PRINC_SURG_PROC_CODE))
        princ_surg_proc_keep = [k for k, v in Counter(tmp).most_common(10)]
        # remove unnecessary variables
        del tmp

        # Load the data
        tx_data = pd.DataFrame()
        for f in files_path:
            df = [
                pd.read_csv(ZipFile(f).open(i), sep="\t", usecols=columns_names)
                for i in ZipFile(f).namelist()
                if "base" in i
            ][0]
            # keep only those rows with one of the 10 most common principal surgical procedure
            df = df[df["PRINC_SURG_PROC_CODE"].isin(princ_surg_proc_keep)]
            # clean up data
            df.dropna(inplace=True)
            df.replace("`", pd.NA, inplace=True)
            df.replace("*", pd.NA, inplace=True)
            # replace sex to numeric
            df.SEX_CODE.replace("M", 0, inplace=True)
            df.SEX_CODE.replace("F", 1, inplace=True)
            df.SEX_CODE.replace("U", 2, inplace=True)
            # set to numerical variable
            for d_code in set(list(df.DISCHARGE_QTR)):
                df.DISCHARGE_QTR.replace(
                    d_code, "".join(d_code.split("Q")), inplace=True
                )
            df.dropna(inplace=True)
            # merge data
            tx_data = pd.concat([tx_data, df])
        # remove uncessary variables
        del df

        # Risk moratality, make it binary
        # 1 Minor
        # 2 Moderate
        # 3 Major
        # 4 Extreme
        tx_data.RISK_MORTALITY.astype(int)
        tx_data.RISK_MORTALITY.replace(1, 0, inplace=True)
        tx_data.RISK_MORTALITY.replace(2, 0, inplace=True)
        tx_data.RISK_MORTALITY.replace(3, 1, inplace=True)
        tx_data.RISK_MORTALITY.replace(4, 1, inplace=True)

        # renumber non-numerical codes for cols
        cols = ["PRINC_DIAG_CODE", "SOURCE_OF_ADMISSION", "E_CODE_1"]
        for col in cols:
            tmp = list(
                {
                    x
                    for x in tx_data[col]
                    if not str(x).isdigit() and not isinstance(x, float)
                }  # pylint: disable=consider-using-set-comprehension
            )
            n = max(
                list(
                    {
                        int(x)
                        for x in tx_data[col]
                        if str(x).isdigit() or isinstance(x, float)
                    }  # pylint: disable=consider-using-set-comprehension
                )
            )
            for i, x in enumerate(tmp):
                tx_data[col].replace(x, n + i, inplace=True)
        del tmp, n
        # set index
        tx_data.set_index("RECORD_ID", inplace=True)
        # final check and drop of NAs
        tx_data.dropna(inplace=True)
        # convert all data to numerical
        tx_data = tx_data.astype(int)
        # save csv file
        tx_data.to_csv(os.path.join(data_folder, "TexasHospitals", processed_data_file))
    else:
        logger.info("Loading processed Texas Hospitals data (2006-2009) csv file.")
        # load texas data processed csv file
        tx_data = pd.read_csv(
            os.path.join(data_folder, "TexasHospitals", processed_data_file)
        )

    # extract target
    var = "RISK_MORTALITY"
    labels = tx_data[var]
    # Drop the column that contains the labels
    tx_data.drop([var], axis=1, inplace=True)

    label_encoder = LabelEncoder()
    encoded_labels = label_encoder.fit_transform(labels.values)
    labels = pd.DataFrame({var: encoded_labels})

    return (tx_data, labels)



"""
Example demonstrating the attribute inference attacks.

Running
-------

Invoke this code from the root AI-SDC folder with
python -m examples.attribute_inference_example

"""
import os

# ignore unused imports because it depends on whether data file is present
from sklearn.datasets import fetch_openml  # pylint:disable=unused-import
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import (  # pylint:disable=unused-import
    LabelEncoder,
    OneHotEncoder,
)

from aisdc.attacks import attribute_attack  # pylint: disable = import-error
from aisdc.attacks.attribute_attack import (
    _get_bounds_risk,
    _infer_categorical,
    _unique_max,
)
from tests.test_attacks_via_safemodel import get_nursery_dataset

# pylint: disable = duplicate-code


def cleanup_file(name: str):
    """removes unwanted files or directory"""
    if os.path.exists(name) and os.path.isfile(name):  # h5
        os.remove(name)


def common_setup():
    """basic commands to get ready to test some code"""
    data = get_nursery_dataset()
    model = RandomForestClassifier(bootstrap=False)
    model.fit(data.x_train, data.y_train)
    attack_args = attribute_attack.AttributeAttackArgs(
        n_cpu=7, report_name="aia_report"
    )

    return data, model, attack_args


def test_attack_args():
    """tests methods in the attack_args class"""
    _, _, attack_args = common_setup()
    _ = attack_args.__str__()  # pylint:disable=unnecessary-dunder-call
    attack_args.set_param("newkey", True)
    thedict = attack_args.get_args()
    assert thedict["newkey"] is True


def test_unique_max():
    """tests the _unique_max helper function"""
    has_unique = (0.3, 0.5, 0.2)
    no_unique = (0.5, 0.5)
    assert _unique_max(has_unique, 0.0) is True
    assert _unique_max(has_unique, 0.6) is False
    assert _unique_max(no_unique, 0.0) is False


def test_categorical_via_modified_attack_brute_force():
    """test lots of functionality for categoricals
    using code from brute_force but without multiprocessing
    """
    data, model, _ = common_setup()

    threshold = 0
    feature = 0
    # make predictions
    _infer_categorical(model, data, feature, threshold)
    # or don't because threshold is too high
    threshold = 999
    _infer_categorical(model, data, feature, threshold)


def test_continuous_via_modified_bounds_risk():
    """tests a lot of the code for continuous variables
    via a copy of the _get_bounds_risk()
    modified not to use multiprocessing
    """
    data, model, _ = common_setup()
    _ = _get_bounds_risk(model, "dummy", 8, data.x_train, data.x_test)


# test below covers a lot of the plotting etc.
def test_AIA_on_nursery():
    """tests running AIA on the nursery data
    with an added continuous feature"""
    data, model, attack_args = common_setup()

    attack_obj = attribute_attack.AttributeAttack(attack_args)
    attack_obj.attack(data, model)

    output = attack_obj.make_report()
    output = output["attack_metrics"]


def test_cleanup():
    """tidies up any files created"""
    files_made = (
        "delete-me.json",
        "aia_example.json",
        "aia_example.pdf",
        "aia_report_cat_frac.png",
        "aia_report_cat_risk.png",
        "aia_report_quant_risk.png",
        "aia_report.pdf",
        "aia_report.json",
    )
    for fname in files_made:
        cleanup_file(fname)



"""test_worst_case_attack.py
Copyright (C) Jim Smith 2022 <james.smith@uwe.ac.uk>
"""
import os
import sys
from unittest.mock import patch

import numpy as np
import pytest
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.svm import SVC

from aisdc.attacks import dataset, worst_case_attack  # pylint: disable = import-error


def clean_up(name):
    """removes unwanted files or directory"""
    if os.path.exists(name) and os.path.isfile(name):
        os.remove(name)


def test_report_worstcase():
    """tests worst case attack directly"""
    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    dataset_obj = dataset.Data()
    dataset_obj.add_processed_data(train_X, train_y, test_X, test_y)

    target_model = SVC(gamma=0.1, probability=True)
    target_model.fit(train_X, train_y)
    _ = target_model.predict_proba(train_X)
    _ = target_model.predict_proba(test_X)

    args = worst_case_attack.WorstCaseAttackArgs(
        # How many attacks to run -- in each the attack model is trained on a different
        # subset of the data
        n_reps=10,
        n_dummy_reps=1,
        p_thresh=0.05,
        in_sample_filename=None,
        out_sample_filename=None,
        test_prop=0.5,
        report_name="test-10reps",
    )

    # with multiple reps
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    attack_obj.attack(dataset_obj, target_model)
    # attack_obj.make_dummy_data() cause exception when used like this!
    _ = attack_obj.make_report()

    # with one rep
    args = worst_case_attack.WorstCaseAttackArgs(
        n_reps=1,
        n_dummy_reps=1,
        p_thresh=0.05,
        in_sample_filename=None,
        out_sample_filename=None,
        test_prop=0.5,
        report_name="test-1rep",
    )

    attack_obj = worst_case_attack.WorstCaseAttack(args)
    attack_obj.attack(dataset_obj, target_model)
    _ = attack_obj.make_report()


def test_attack_with_correct_feature():
    """Test the attack when the model correctness feature is used"""
    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    dataset_obj = dataset.Data()
    dataset_obj.add_processed_data(train_X, train_y, test_X, test_y)

    target_model = SVC(gamma=0.1, probability=True)
    target_model.fit(train_X, train_y)

    args = worst_case_attack.WorstCaseAttackArgs(
        # How many attacks to run -- in each the attack model is trained on a different
        # subset of the data
        n_reps=1,
        n_dummy_reps=1,
        p_thresh=0.05,
        in_sample_filename=None,
        out_sample_filename=None,
        test_prop=0.5,
        report_name="test-1rep",
        include_model_correct_feature=True,
    )

    # with multiple reps
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    attack_obj.attack(dataset_obj, target_model)

    # Check that attack_metrics has the Yeom metrics
    assert "yeom_tpr" in attack_obj.attack_metrics[0]
    assert "yeom_fpr" in attack_obj.attack_metrics[0]
    assert "yeom_advantage" in attack_obj.attack_metrics[0]


def test_attack_from_predictions():
    """checks code that runs attacks from predictions"""

    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    dataset_obj = dataset.Data()
    dataset_obj.add_processed_data(train_X, train_y, test_X, test_y)

    target_model = SVC(gamma=0.1, probability=True)
    target_model.fit(train_X, train_y)
    ytr_pred = target_model.predict_proba(train_X)
    yte_pred = target_model.predict_proba(test_X)
    np.savetxt("ypred_train.csv", ytr_pred, delimiter=",")
    np.savetxt("ypred_test.csv", yte_pred, delimiter=",")

    args = worst_case_attack.WorstCaseAttackArgs(
        # How many attacks to run -- in each the attack model is trained on a different
        # subset of the data
        n_reps=10,
        n_dummy_reps=1,
        p_thresh=0.05,
        in_sample_filename="ypred_train.csv",
        out_sample_filename="ypred_test.csv",
        test_prop=0.5,
        report_name="test-10reps",
    )

    assert args.get_args()["in_sample_filename"] == "ypred_train.csv"
    print(args)

    # with multiple reps
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    attack_obj.attack_from_prediction_files()


def test_attack_from_predictions_no_dummy():
    """checks code that runs attacks from predictions"""

    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    dataset_obj = dataset.Data()
    dataset_obj.add_processed_data(train_X, train_y, test_X, test_y)

    target_model = SVC(gamma=0.1, probability=True)
    target_model.fit(train_X, train_y)
    ytr_pred = target_model.predict_proba(train_X)
    yte_pred = target_model.predict_proba(test_X)
    np.savetxt("ypred_train.csv", ytr_pred, delimiter=",")
    np.savetxt("ypred_test.csv", yte_pred, delimiter=",")

    args = worst_case_attack.WorstCaseAttackArgs(
        # How many attacks to run -- in each the attack model is trained on a different
        # subset of the data
        n_reps=10,
        n_dummy_reps=0,
        p_thresh=0.05,
        in_sample_filename="ypred_train.csv",
        out_sample_filename="ypred_test.csv",
        test_prop=0.5,
        report_name="test-10reps",
    )

    assert args.get_args()["in_sample_filename"] == "ypred_train.csv"
    print(args)

    # with multiple reps
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    attack_obj.attack_from_prediction_files()


def test_dummy_data():
    """test functionality around creating dummy data"""
    args = worst_case_attack.WorstCaseAttackArgs(
        # How many attacks to run -- in each the attack model is trained on a different
        # subset of the data
        n_reps=10,
        n_dummy_reps=1,
        p_thresh=0.05,
        in_sample_filename="ypred_train.csv",
        out_sample_filename="ypred_test.csv",
        test_prop=0.5,
        report_name="test-10reps",
    )

    attack_obj = worst_case_attack.WorstCaseAttack(args)
    attack_obj.make_dummy_data()


def test_attack_data_prep():
    """test the method that prepares the attack data"""
    args = worst_case_attack.WorstCaseAttackArgs()
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    train_preds = np.array([[1, 0], [0, 1]], int)
    test_preds = np.array([[2, 0], [0, 2]], int)

    mi_x, mi_y = attack_obj._prepare_attack_data(  # pylint: disable=protected-access
        train_preds, test_preds
    )
    np.testing.assert_array_equal(mi_y, np.array([1, 1, 0, 0], np.int))
    # Test the x data produced. Each row should be sorted in descending order
    np.testing.assert_array_equal(mi_x, np.array([[1, 0], [1, 0], [2, 0], [2, 0]]))

    # With sort_probs = False, the rows of x should not be sorted
    args = worst_case_attack.WorstCaseAttackArgs(sort_probs=False)
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    mi_x, mi_y = attack_obj._prepare_attack_data(  # pylint: disable=protected-access
        train_preds, test_preds
    )
    np.testing.assert_array_equal(mi_y, np.array([1, 1, 0, 0], np.int))
    np.testing.assert_array_equal(mi_x, np.array([[1, 0], [0, 1], [2, 0], [0, 2]]))


def test_attack_data_prep_with_correct_feature():
    """test the method that prepares the attack data.
    This time, testing that the model correctness values are added, are always
    the final feature, and are not included in the sorting"""
    args = worst_case_attack.WorstCaseAttackArgs(include_model_correct_feature=True)
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    train_preds = np.array([[1, 0], [0, 1]], int)
    test_preds = np.array([[2, 0], [0, 2]], int)
    train_correct = np.array([1, 0], int)
    test_correct = np.array([0, 1], int)

    mi_x, mi_y = attack_obj._prepare_attack_data(  # pylint: disable=protected-access
        train_preds, test_preds, train_correct=train_correct, test_correct=test_correct
    )
    np.testing.assert_array_equal(mi_y, np.array([1, 1, 0, 0], np.int))
    # Test the x data produced. Each row should be sorted in descending order
    np.testing.assert_array_equal(
        mi_x, np.array([[1, 0, 1], [1, 0, 0], [2, 0, 0], [2, 0, 1]])
    )

    # With sort_probs = False, the rows of x should not be sorted
    args = worst_case_attack.WorstCaseAttackArgs(
        sort_probs=False, include_model_correct_feature=True
    )
    attack_obj = worst_case_attack.WorstCaseAttack(args)
    mi_x, mi_y = attack_obj._prepare_attack_data(  # pylint: disable=protected-access
        train_preds, test_preds, train_correct=train_correct, test_correct=test_correct
    )
    np.testing.assert_array_equal(mi_y, np.array([1, 1, 0, 0], np.int))
    np.testing.assert_array_equal(
        mi_x, np.array([[1, 0, 1], [0, 1, 0], [2, 0, 0], [0, 2, 1]])
    )


def test_non_rf_mia():
    """Tests that it is possible to set the attack model via the args
    In this case, we set as a SVC. But we set probability to false. If the code does
    indeed try and use the SVC (as we want) it will fail as it will try and access
    the predict_proba which won't work if probability=False. Hence, if the code throws
    an AttributeError we now it *is* trying to use the SVC"""

    args = worst_case_attack.WorstCaseAttackArgs(
        mia_attack_model=SVC,
        mia_attack_model_hyp={"kernel": "rbf", "probability": False},
    )

    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)
    dataset_obj = dataset.Data()
    dataset_obj.add_processed_data(train_X, train_y, test_X, test_y)

    target_model = SVC(gamma=0.1, probability=True)
    target_model.fit(train_X, train_y)
    ytr_pred = target_model.predict_proba(train_X)
    yte_pred = target_model.predict_proba(test_X)

    attack_obj = worst_case_attack.WorstCaseAttack(args)
    with pytest.raises(AttributeError):
        attack_obj.attack_from_preds(ytr_pred, yte_pred)


def test_main():
    """test invocation via command line"""

    # option 1
    testargs = ["prog", "make-dummy-data"]
    with patch.object(sys, "argv", testargs):
        worst_case_attack.main()

    # option 2
    testargs = ["prog", "run-attack"]
    with patch.object(sys, "argv", testargs):
        worst_case_attack.main()

    # wrong args

    # testargs = ["prog", "run-attack","--no-such-arg"]
    # with patch.object(sys, 'argv', testargs):
    #    worst_case_attack.main()


def test_cleanup():
    """gets rid of files created during tests"""
    names = [
        "worstcase_report.pdf",
        "log_roc.png",
        "worstcase_report.json",
        "test_preds.csv",
        "train_preds.csv",
        "ypred_test.csv",
        "ypred_train.csv",
        "test-1rep.pdf",
        "test-1rep.json",
        "test-10reps.pdf",
        "test-10reps.json",
    ]
    for name in names:
        clean_up(name)



"""makes class for various models supported"""
from .dp_svc import DPSVC
from .safedecisiontreeclassifier import SafeDecisionTreeClassifier
from .safekeras import SafeKerasModel
from .saferandomforestclassifier import SafeRandomForestClassifier
from .safesvc import SafeSVC
from .safetf import Safe_tf_DPModel



"""
Test the data interface code
"""

import unittest

import pandas as pd

from aisdc.preprocessing.loaders import (  # pylint: disable = import-error
    UnknownDataset,
    get_data_sklearn,
)


class TestLoaders(unittest.TestCase):
    """
    Test the data loaders
    """

    def test_iris(self):
        """Nursery data"""
        feature_df, target_df = get_data_sklearn("iris")
        self.assertIsInstance(feature_df, pd.DataFrame)
        self.assertIsInstance(target_df, pd.DataFrame)

    def test_unknown(self):
        """
        Test that a nonsense string raises the correct exception
        """
        with self.assertRaises(UnknownDataset):
            _, _ = get_data_sklearn("NONSENSE")

    def test_standard(self):
        """
        Test that standardisation creates standard features
        """
        feature_df, _ = get_data_sklearn("standard iris")
        for column in feature_df.columns:
            temp = feature_df[column].mean()
            self.assertAlmostEqual(temp, 0.0)
            temp = feature_df[column].std()
            self.assertAlmostEqual(temp, 1.0)

    def test_minmax(self):
        """
        Test the minmax scaling
        """
        feature_df, _ = get_data_sklearn("minmax iris")
        for column in feature_df.columns:
            temp = feature_df[column].min()
            self.assertAlmostEqual(temp, 0.0)
            temp = feature_df[column].max()
            self.assertAlmostEqual(temp, 1.0)



"""
worst_case_attack.py

Runs a worst case attack based upon predictive probabilities stored in two .csv files
"""

from __future__ import annotations

import argparse
import logging
from collections.abc import Hashable
from typing import Any

import numpy as np
import sklearn
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split

from aisdc.attacks import metrics, report
from aisdc.attacks.attack import Attack
from aisdc.attacks.dataset import Data

logging.basicConfig(level=logging.INFO)

P_THRESH = 0.05


class WorstCaseAttackArgs:
    """Arguments for worst case"""

    def __init__(self, **kwargs):
        self.__dict__["n_reps"] = 10
        self.__dict__["p_thresh"] = 0.05
        self.__dict__["n_dummy_reps"] = 1
        self.__dict__["train_beta"] = 2
        self.__dict__["test_beta"] = 2
        self.__dict__["test_prop"] = 0.3
        self.__dict__["n_rows_in"] = 1000
        self.__dict__["n_rows_out"] = 1000
        self.__dict__["in_sample_filename"] = None
        self.__dict__["out_sample_filename"] = None
        self.__dict__["report_name"] = None
        self.__dict__["include_model_correct_feature"] = False
        self.__dict__["sort_probs"] = True
        self.__dict__["mia_attack_model"] = RandomForestClassifier
        self.__dict__["mia_attack_model_hyp"] = {
            "min_samples_split": 20,
            "min_samples_leaf": 10,
            "max_depth": 5,
        }
        self.__dict__.update(kwargs)

    def __str__(self):
        return ",".join(
            [f"{str(key)}: {str(value)}" for key, value in self.__dict__.items()]
        )

    def set_param(self, key: Hashable, value: Any) -> None:
        """Set a parameter"""
        self.__dict__[key] = value

    def get_args(self) -> dict:
        """Return arguments"""
        return self.__dict__


class WorstCaseAttack(Attack):
    """Class to wrap the worst case attack code"""

    def __init__(self, args: WorstCaseAttackArgs = WorstCaseAttackArgs()):
        self.attack_metrics = None
        self.dummy_attack_metrics = None
        self.metadata = None
        self.args = args

    def __str__(self):
        return "WorstCase attack"

    def attack(self, dataset: Data, target_model: sklearn.base.BaseEstimator) -> None:
        """Programmatic attack entry point

        To be used when code has access to data class and trained target model

        Parameters
        ----------
        dataset: attacks.dataset.Data
            dataset as a Data class object
        target_model: sklearn.base.BaseEstimator
            target model that inherits from an sklearn BaseEstimator
        """
        train_preds = target_model.predict_proba(dataset.x_train)
        test_preds = target_model.predict_proba(dataset.x_test)
        train_correct = None
        test_correct = None
        if self.args.include_model_correct_feature:
            train_correct = 1 * (
                dataset.y_train == target_model.predict(dataset.x_train)
            )
            test_correct = 1 * (dataset.y_test == target_model.predict(dataset.x_test))

        self.attack_from_preds(
            train_preds,
            test_preds,
            train_correct=train_correct,
            test_correct=test_correct,
        )

    def attack_from_prediction_files(self):
        """Start an attack from saved prediction files

        To be used when only saved predictions are available.

        Filenames for the saved prediction files to be specified in the arguments provided
        in the constructor
        """
        train_preds = np.loadtxt(self.args.in_sample_filename, delimiter=",")
        test_preds = np.loadtxt(self.args.out_sample_filename, delimiter=",")
        self.attack_from_preds(train_preds, test_preds)

    def attack_from_preds(  # pylint: disable=too-many-locals
        self,
        train_preds: np.ndarray,
        test_preds: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> None:
        """
        Runs the attack based upon the predictions in train_preds and test_preds, and the params
        stored in self.args

        Parameters
        ----------

        train_preds: np.ndarray
            Array of train predictions. One row per example, one column per class (i.e. 2)
        test_preds:  np.ndarray
            Array of test predictions. One row per example, one column per class (i.e. 2)

        """
        logger = logging.getLogger("attack-from-preds")
        logger.info("Running main attack repetitions")
        self.attack_metrics = self.run_attack_reps(
            train_preds,
            test_preds,
            train_correct=train_correct,
            test_correct=test_correct,
        )

        if self.args.n_dummy_reps > 0:
            logger.info("Running dummy attack reps")
            self.dummy_attack_metrics = []
            n_train_rows = len(train_preds)
            n_test_rows = len(test_preds)
            for _ in range(self.args.n_dummy_reps):
                d_train_preds, d_test_preds = self.generate_arrays(
                    n_train_rows, n_test_rows, self.args.train_beta, self.args.test_beta
                )
                temp_metrics = self.run_attack_reps(d_train_preds, d_test_preds)
                self.dummy_attack_metrics += temp_metrics
        logger.info("Finished running attacks")

    def _prepare_attack_data(
        self,
        train_preds: np.ndarray,
        test_preds: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Prepare training data and labels for attack model
        Combines the train and test preds into a single numpy array (optionally) sorting each
        row to have the highest probabilities in the first column. Constructs a label array that
        has ones corresponding to training rows and zeros to testing rows."""
        logger = logging.getLogger("prep-attack-data")
        if self.args.sort_probs:
            logger.info("Sorting probabilities to leave highest value in first column")
            train_preds = -np.sort(-train_preds, axis=1)
            test_preds = -np.sort(-test_preds, axis=1)

        logger.info("Creating MIA data")

        if self.args.include_model_correct_feature and train_correct is not None:
            train_preds = np.hstack((train_preds, train_correct[:, None]))
            test_preds = np.hstack((test_preds, test_correct[:, None]))

        mi_x = np.vstack((train_preds, test_preds))
        mi_y = np.hstack((np.ones(len(train_preds)), np.zeros(len(test_preds))))

        return (mi_x, mi_y)

    def run_attack_reps(  # pylint: disable = too-many-locals
        self,
        train_preds: np.ndarray,
        test_preds: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> list:
        """
        Run actual attack reps from train and test predictions

        Parameters
        ----------
        train_preds: np.ndarray
            predictions from the model on training (in-sample) data
        test_preds: np.ndarray
            predictions from the model on testing (out-of-sample) data

        Returns
        -------
        mia_metrics: List
            List of attack metrics dictionaries, one for each repetition
        """
        self.args.set_param("n_rows_in", len(train_preds))
        self.args.set_param("n_rows_out", len(test_preds))
        logger = logging.getLogger("attack-reps")
        mi_x, mi_y = self._prepare_attack_data(
            train_preds, test_preds, train_correct, test_correct
        )

        mia_metrics = []
        for rep in range(self.args.n_reps):
            logger.info("Rep %d of %d", rep, self.args.n_reps)
            mi_train_x, mi_test_x, mi_train_y, mi_test_y = train_test_split(
                mi_x, mi_y, test_size=self.args.test_prop, stratify=mi_y
            )
            attack_classifier = self.args.mia_attack_model(
                **self.args.mia_attack_model_hyp
            )
            attack_classifier.fit(mi_train_x, mi_train_y)

            mia_metrics.append(
                metrics.get_metrics(attack_classifier, mi_test_x, mi_test_y)
            )

            if self.args.include_model_correct_feature and train_correct is not None:
                # Compute the Yeom TPR and FPR
                yeom_preds = mi_test_x[:, -1]
                tn, fp, fn, tp = confusion_matrix(mi_test_y, yeom_preds).ravel()
                mia_metrics[-1]["yeom_tpr"] = tp / (tp + fn)
                mia_metrics[-1]["yeom_fpr"] = fp / (fp + tn)
                mia_metrics[-1]["yeom_advantage"] = (
                    mia_metrics[-1]["yeom_tpr"] - mia_metrics[-1]["yeom_fpr"]
                )

        logger.info("Finished simulating attacks")

        return mia_metrics

    def _get_global_metrics(self, attack_metrics: list) -> dict:
        """Summarise metrics from a metric list

        Arguments
        ---------
        attack_metrics: List
            list of attack metrics dictionaries

        Returns
        -------
        global_metrics: Dict
            Dictionary of summary metrics

        """
        global_metrics = {}
        auc_p_vals = [
            metrics.auc_p_val(
                m["AUC"], m["n_pos_test_examples"], m["n_neg_test_examples"]
            )[0]
            for m in attack_metrics
        ]

        m = attack_metrics[0]
        _, auc_std = metrics.auc_p_val(
            0.5, m["n_pos_test_examples"], m["n_neg_test_examples"]
        )

        global_metrics[
            "null_auc_3sd_range"
        ] = f"{0.5 - 3*auc_std:.4f} -> {0.5 + 3*auc_std:.4f}"
        global_metrics["n_sig_auc_p_vals"] = self._get_n_significant(
            auc_p_vals, self.args.p_thresh
        )
        global_metrics["n_sig_auc_p_vals_corrected"] = self._get_n_significant(
            auc_p_vals, self.args.p_thresh, bh_fdr_correction=True
        )

        pdif_vals = [np.exp(-m["PDIF01"]) for m in attack_metrics]
        global_metrics["n_sig_pdif_vals"] = self._get_n_significant(
            pdif_vals, self.args.p_thresh
        )
        global_metrics["n_sig_pdif_vals_corrected"] = self._get_n_significant(
            pdif_vals, self.args.p_thresh, bh_fdr_correction=True
        )

        return global_metrics

    def _get_n_significant(self, p_val_list, p_thresh, bh_fdr_correction=False):
        """
        Helper method to determine if values within a list of p-values are significant at
        p_thresh. Can perform multiple testing correction.
        """
        if not bh_fdr_correction:
            return sum(
                1 for p in p_val_list if p <= p_thresh
            )  # pylint: disable = consider-using-generator
        p_val_list = np.asarray(sorted(p_val_list))
        n_vals = len(p_val_list)
        hoch_vals = np.array([(k / n_vals) * P_THRESH for k in range(1, n_vals + 1)])
        bh_sig_list = p_val_list <= hoch_vals
        if any(bh_sig_list):
            n_sig_bh = (np.where(bh_sig_list)[0]).max() + 1
        else:
            n_sig_bh = 0
        return n_sig_bh

    def _generate_array(self, n_rows: int, beta: float) -> np.ndarray:
        """Generate a single array of predictions, used when doing baseline experiments

        Parameters
        ----------
        n_rows: int
            the number of rows worth of data to generate
        beta: float
            the beta parameter for sampling probabilities

        Returns
        -------
        preds: np.ndarray
            Array of predictions. Two columns, n_rows rows

        Notes
        -----

        Examples
        --------

        """

        preds = np.zeros((n_rows, 2), float)
        for row_idx in range(n_rows):
            train_class = np.random.choice(2)
            train_prob = np.random.beta(1, beta)
            preds[row_idx, train_class] = train_prob
            preds[row_idx, 1 - train_class] = 1 - train_prob
        return preds

    def generate_arrays(
        self,
        n_rows_in: int,
        n_rows_out: int,
        train_beta: float = 2,
        test_beta: float = 2,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Generate train and test prediction arrays, used when computing baseline

        Parameters
        ----------
        n_rows_in: int
            number of rows of in-sample (training) probabilities
        n_rows_out: int
            number of rows of out-of-sample (testing) probabilities
        train_beta: float
            beta value for generating train probabilities
        test_beta: float:
            beta_value for generating test probabilities

        Returns
        -------
        train_preds: np.ndarray
            Array of train predictions (n_rows x 2 columns)
        test_preds: np.ndarray
            Array of test predictions (n_rows x 2 columns)
        """
        train_preds = self._generate_array(n_rows_in, train_beta)
        test_preds = self._generate_array(n_rows_out, test_beta)
        return train_preds, test_preds

    def make_dummy_data(self) -> None:
        """Makes dummy data for testing functionality

        Parameters
        ----------
        args: dict
            Command line arguments

        Returns
        -------

        Notes
        -----

        Returns nothing but saves two .csv files
        """
        logger = logging.getLogger("dummy-data")
        logger.info(
            "Making dummy data with %d rows in and %d out",
            self.args.n_rows_in,
            self.args.n_rows_out,
        )
        logger.info("Generating rows")
        train_preds, test_preds = self.generate_arrays(
            self.args.n_rows_in,
            self.args.n_rows_out,
            train_beta=self.args.train_beta,
            test_beta=self.args.test_beta,
        )
        logger.info("Saving files")
        np.savetxt(self.args.in_sample_filename, train_preds, delimiter=",")
        np.savetxt(self.args.out_sample_filename, test_preds, delimiter=",")

    def _construct_metadata(self):
        """Constructs the metadata object, after attacks"""
        self.metadata = {}
        # Store all args
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"].update(self.args.__dict__)
        if "func" in self.metadata["experiment_details"]:
            del self.metadata["experiment_details"]["func"]

        self.metadata["attack"] = str(self)

        # Global metrics
        self.metadata["global_metrics"] = self._get_global_metrics(self.attack_metrics)
        self.metadata["baseline_global_metrics"] = self._get_global_metrics(
            self.dummy_attack_metrics
        )

    def make_report(self) -> dict:
        """Creates output dictionary structure"""
        output = {}
        output["attack_metrics"] = self.attack_metrics
        output["dummy_attack_metrics"] = self.dummy_attack_metrics
        self._construct_metadata()
        output["metadata"] = self.metadata
        if self.args.report_name is not None:
            json_report = report.create_json_report(output)
            with open(f"{self.args.report_name}.json", "w", encoding="utf-8") as f:
                f.write(json_report)

            pdf = report.create_mia_report(output)
            pdf.output(f"{self.args.report_name}.pdf", "F")

        return output


def _make_dummy_data(args):
    """Initialise class and run dummy data creation"""
    wc_args = WorstCaseAttackArgs(**args.__dict__)
    wc_args.set_param("in_sample_filename", "train_preds.csv")
    wc_args.set_param("out_sample_filename", "test_preds.csv")
    attack_obj = WorstCaseAttack(wc_args)
    attack_obj.make_dummy_data()


def _run_attack(args):
    """Initialise class and run attack from prediction files"""
    wc_args = WorstCaseAttackArgs(**args.__dict__)
    attack_obj = WorstCaseAttack(wc_args)
    attack_obj.attack_from_prediction_files()
    _ = attack_obj.make_report()


def main():
    """main method to parse arguments and invoke relevant method"""
    logger = logging.getLogger("main")
    parser = argparse.ArgumentParser(
        description=("Perform a worst case attack from saved model predictions")
    )

    subparsers = parser.add_subparsers()
    dummy_parser = subparsers.add_parser("make-dummy-data")
    dummy_parser.add_argument(
        "--num-rows-in",
        action="store",
        dest="n_rows_in",
        type=int,
        required=False,
        default=1000,
        help=("How many rows to generate in the in-sample file. Default = %(default)d"),
    )

    dummy_parser.add_argument(
        "--num-rows-out",
        action="store",
        dest="n_rows_out",
        type=int,
        required=False,
        default=1000,
        help=(
            "How many rows to generate in the out-of-sample file. Default = %(default)d"
        ),
    )

    dummy_parser.add_argument(
        "--train-beta",
        action="store",
        type=float,
        required=False,
        default=5,
        dest="train_beta",
        help=(
            "Value of b parameter for beta distribution used to sample the in-sample "
            "probabilities. "
            "High values will give more extreme probabilities. Set this value higher than "
            "--test-beta to see successful attacks. Default = %(default)f"
        ),
    )

    dummy_parser.add_argument(
        "--test-beta",
        action="store",
        type=float,
        required=False,
        default=2,
        dest="test_beta",
        help=(
            "Value of b parameter for beta distribution used to sample the out-of-sample "
            "probabilities. High values will give more extreme probabilities. Set this value "
            "lower than --train-beta to see successful attacks. Default = %(default)f"
        ),
    )

    dummy_parser.set_defaults(func=_make_dummy_data)

    attack_parser = subparsers.add_parser("run-attack")
    attack_parser.add_argument(
        "-i",
        "--in-sample-preds",
        action="store",
        dest="in_sample_filename",
        required=False,
        type=str,
        default="train_preds.csv",
        help=(
            "csv file containing the predictive probabilities (one column per class) for the "
            "training data (one row per training example). Default = %(default)s"
        ),
    )

    attack_parser.add_argument(
        "-o",
        "--out-of-sample-preds",
        action="store",
        dest="out_sample_filename",
        required=False,
        type=str,
        default="test_preds.csv",
        help=(
            "csv file containing the predictive probabilities (one column per class) for the "
            "non-training data (one row per training example). Default = %(default)s"
        ),
    )

    attack_parser.add_argument(
        "-r",
        "--n-reps",
        type=int,
        required=False,
        default=5,
        action="store",
        dest="n_reps",
        help=(
            "Number of repetitions (splitting data into attack model training and testing "
            "partitions to perform. Default = %(default)d"
        ),
    )

    attack_parser.add_argument(
        "-t",
        "--test-prop",
        type=float,
        required=False,
        default=0.3,
        action="store",
        dest="test_prop",
        help=(
            "Proportion of examples to be used for testing when fiting the attack model. "
            "Default = %(default)f"
        ),
    )

    attack_parser.add_argument(
        "--report-name",
        type=str,
        action="store",
        dest="report_name",
        default="worstcase_report",
        required=False,
        help=(
            "Filename for the report output. Default = %(default)s. Code will append .pdf and .json"
        ),
    )

    attack_parser.add_argument(
        "--n-dummy-reps",
        type=int,
        action="store",
        dest="n_dummy_reps",
        default=1,
        required=False,
        help=(
            "Number of dummy datasets to sample. Each will be assessed with --n-reps train and "
            "test splits. Set to 0 to do no baseline calculations. Default = %(default)d"
        ),
    )

    attack_parser.add_argument(
        "--p-thresh",
        action="store",
        type=float,
        default=P_THRESH,
        required=False,
        dest="p_thresh",
        help=("P-value threshold for significance testing. Default = %(default)f"),
    )

    # Not currently possible from the command line as we cannot compute the correctness
    # of predictions. Possibly to be added in the future
    # attack_parser.add_argument(
    #     "--include-correct",
    #     action="store",
    #     type=bool,
    #     required=False,
    #     default=False,
    #     dest='include_model_correct_feature',
    #     help=(
    #         "Whether or not to include an additional feature into the MIA attack model that "
    #         "holds whether or not the target model made a correct predicion for each example."
    #     ),
    # )

    attack_parser.add_argument(
        "--sort-probs",
        action="store",
        type=bool,
        default=True,
        required=False,
        dest="sort_probs",
        help=(
            "Whether or not to sort the output probabilities (per row) before "
            "using them to train the attack model. Default = %(default)f"
        ),
    )

    attack_parser.set_defaults(func=_run_attack)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        logger.error("Invalid command. Try --help to get more details")
        logger.error(e)


if __name__ == "__main__":  # pragma:no cover
    main()



""" work in progress to allow use of the DPModel classes
 Jim smith 2022
 When ready, linting of  the imports will be enabled
"""


# pylint: disable=unused-import
from typing import Any

import tensorflow as tf
import tensorflow_privacy as tf_privacy
from tensorflow_privacy import DPModel

# from tensorflow.keras import Model as KerasModel
from tensorflow_privacy.privacy.analysis import compute_dp_sgd_privacy
from tensorflow_privacy.privacy.optimizers import dp_optimizer_keras

from ..safemodel import SafeModel


class Safe_tf_DPModel(SafeModel, DPModel):
    """Privacy Protected tensorflow_privacy DP-SGD subclass of Keras model"""

    # remove comment once model starts to be populated
    # pylint:disable=super-init-not-called
    def __init__(
        self,
        l2_norm_clip: float,
        noise_multiplier: float,
        use_xla: bool,
        *args: any,
        **kwargs: any
    ) -> None:
        """creates model and applies constraints to parameters"""
        # safemodel.__init__(self)
        # DPModel.__init__(self, **kwargs)
        # self.model_type: str = "tf_DPModel"
        # super().preliminary_check(apply_constraints=True, verbose=True)
        raise NotImplementedError



"""Jim Smith October 2022
tests to pick up odd cases not otherwise covered
in code in the attacks folder
"""
import math

import numpy as np
import pytest
from fpdf import FPDF

from aisdc.attacks import attack, dataset, mia_extremecase, report
from aisdc.safemodel.classifiers import SafeDecisionTreeClassifier

BORDER = 0


def test_superclass():
    """Test that the exceptions are raised if the superclass is called in error"""
    my_attack = attack.Attack()
    dataset_obj = dataset.Data()
    target_obj = SafeDecisionTreeClassifier()
    with pytest.raises(NotImplementedError):
        my_attack.attack(target_obj, dataset_obj)
    with pytest.raises(NotImplementedError):
        print(str(my_attack))  # .__str__()


def test_mia_extremecase():
    """test the extreme case mia in the file of the same name"""

    # create actual values
    y = np.zeros(50000)
    y[:25] = 1
    # exactly right and wrong predictions
    right = np.zeros(50000)
    right[:25] = 1
    wrong = 1 - right

    # right predictions - triggers override for very small logp
    _, _, _, pval = mia_extremecase.min_max_disc(y, right)
    assert pval == -115.13

    # wrong predictions - probaility very close to 1 so logp=0
    _, _, _, pval = mia_extremecase.min_max_disc(y, wrong)
    assert math.isclose(pval, 0.0)


def test_NumpyArrayEncoder():
    """conversion routine
    from reports.py
    """

    i32 = np.int32(2)
    i64 = np.int64(2)
    twoDarray = np.zeros((2, 2))
    my_encoder = report.NumpyArrayEncoder()

    retval = my_encoder.default(i32)
    assert isinstance(retval, int)

    retval = my_encoder.default(i64)
    assert isinstance(retval, int)

    retval = my_encoder.default(twoDarray)
    assert isinstance(retval, list)

    with pytest.raises(TypeError):
        retval = my_encoder.default("a string")


def test_line():
    """code from report.py"""
    pdf = FPDF()
    pdf.add_page()
    report.line(pdf, "foo")
    pdf.close()


def test_dict():
    """code from report.py"""
    pdf = FPDF()
    pdf.add_page()
    mydict = {"a": "hello", "b": "world"}
    report._write_dict(  # pylint:disable=protected-access
        pdf, mydict, indent=0, border=BORDER
    )
    pdf.close()



"""
Tests attacks called via safemodel classes
uses a subsampled nursery dataset as this tests more of the attack code
currently using random forests
"""
import os
import pickle
from pathlib import Path

import numpy as np
from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

from aisdc.attacks import attribute_attack, likelihood_attack, worst_case_attack
from aisdc.attacks.dataset import Data
from aisdc.safemodel.classifiers import SafeDecisionTreeClassifier

# pylint: disable=too-many-locals,bare-except,duplicate-code,unnecessary-dunder-call


def cleanup_file(name: str):
    """removes unwanted files or directory"""
    if os.path.exists(name) and os.path.isfile(name):  # h5
        os.remove(name)


def get_nursery_dataset() -> Data:
    """returns a randomly sampled 10+10% of
    the nursery data set as a Data object
    if needed fetches it from openml and saves. it

    """

    the_file = os.path.normpath("tests/datasets/nursery_as_dataset.pickle")
    save_locally = True
    need_download = True

    if Path(the_file).is_file():
        try:
            with open(the_file, "rb") as f:
                the_data = pickle.load(f)
            need_download = False
        except:  # pragma: no cover
            need_download = True

    if need_download:  # pragma: no cover
        nursery_data = fetch_openml(data_id=26, as_frame=True)
        x = np.asarray(nursery_data.data, dtype=str)
        y = np.asarray(nursery_data.target, dtype=str)
        # change labels from recommend to priority for the two odd cases
        num = len(y)
        for i in range(num):
            if y[i] == "recommend":
                y[i] = "priority"

        indices: list[list[int]] = [
            [0, 1, 2],  # parents
            [3, 4, 5, 6, 7],  # has_nurs
            [8, 9, 10, 11],  # form
            [12, 13, 14, 15],  # children
            [16, 17, 18],  # housing
            [19, 20],  # finance
            [21, 22, 23],  # social
            [24, 25, 26],  # health
            [27],  # dummy
        ]

        # [Researcher] Split into training and test sets
        # target model train / test split - these are strings
        (
            x_train_orig,
            x_test_orig,
            y_train_orig,
            y_test_orig,
        ) = train_test_split(
            x,
            y,
            test_size=0.05,
            stratify=y,
            shuffle=True,
        )

        # now resample the training data reduce number of examples
        _, x_train_orig, _, y_train_orig = train_test_split(
            x_train_orig,
            y_train_orig,
            test_size=0.05,
            stratify=y_train_orig,
            shuffle=True,
        )

        # [Researcher] Preprocess dataset
        # one-hot encoding of features and integer encoding of labels
        label_enc = LabelEncoder()
        feature_enc = OneHotEncoder()
        x_train = feature_enc.fit_transform(x_train_orig).toarray()
        y_train = label_enc.fit_transform(y_train_orig)
        x_test = feature_enc.transform(x_test_orig).toarray()
        y_test = label_enc.transform(y_test_orig)

        # add dummy continuous valued attribute from N(0.5,0.05)
        dummy_tr = 0.5 + 0.05 * np.random.randn(x_train.shape[0])
        dummy_te = 0.5 + 0.05 * np.random.randn(x_test.shape[0])
        dummy_all = np.hstack((dummy_tr, dummy_te)).reshape(-1, 1)
        dummy_tr = dummy_tr.reshape(-1, 1)
        dummy_te = dummy_te.reshape(-1, 1)

        x_train = np.hstack((x_train, dummy_tr))
        x_train_orig = np.hstack((x_train_orig, dummy_tr))
        x_test = np.hstack((x_test, dummy_te))
        x_test_orig = np.hstack((x_test_orig, dummy_te))
        xmore = np.concatenate((x_train_orig, x_test_orig))
        n_features = np.shape(x_train_orig)[1]

        # [TRE / Researcher] Wrap the data in a dataset object
        the_data = Data()
        the_data.name = "nursery"
        the_data.add_processed_data(x_train, y_train, x_test, y_test)
        for i in range(n_features - 1):
            the_data.add_feature(nursery_data.feature_names[i], indices[i], "onehot")
        the_data.add_feature("dummy", indices[n_features - 1], "float")
        the_data.add_raw_data(
            xmore, y, x_train_orig, y_train_orig, x_test_orig, y_test_orig
        )

    if need_download and save_locally:  # pragma:no cover
        # make directory if needed then save
        # tested in local versions but don't want this in CI testing
        output_file = Path(the_file)
        output_file.parent.mkdir(exist_ok=True, parents=True)
        with open(the_file, "wb") as f:
            pickle.dump(the_data, f)

    return the_data


def test_attacks_via_request_release():
    """make vulnerable,hacked model then call request_release"""
    the_data = get_nursery_dataset()
    assert the_data.__str__() == "nursery"

    # build a broken model and hack it so lots of reasons to fail and be vulnerable
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=10, min_samples_leaf=1)
    model.fit(the_data.x_train, the_data.y_train)
    model.min_samples_leaf = 10
    model.request_release(filename="vulnerable_hacked.pkl", data_obj=the_data)
    files_made = (
        "vulnerable_hacked_attribute_res.json",
        "vulnerable_hacked_lira_res.json",
        "vulnerable_hacked_worst_case_res.json",
        "vulnerable_hacked_data.pickle",
        "vulnerable_hacked.pkl",
    )
    for fname in files_made:
        cleanup_file(fname)


def test_run_attack_lira():
    """calls the lira attack via safemodel"""
    the_data = get_nursery_dataset()
    assert the_data.__str__() == "nursery"

    # build a model
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5)
    model.fit(the_data.x_train, the_data.y_train)
    _, disclosive = model.preliminary_check()
    assert not disclosive

    print(np.unique(the_data.y_test, return_counts=True))

    print(np.unique(model.predict(the_data.x_test), return_counts=True))

    fname = "delete-me"
    metadata = model.run_attack(the_data, "lira", fname)
    files_made = (
        "delete-me.json",
        "lira_example_report.json",
        "lira_example_report.pdf",
        "log_roc.png",
    )
    for fname in files_made:
        cleanup_file(fname)
    assert len(metadata) > 0  # something has been added


def test_run_attack_worstcase():
    """calls the worst case attack via safemodel"""
    the_data = get_nursery_dataset()
    assert the_data.__str__() == "nursery"

    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5)
    model.fit(the_data.x_train, the_data.y_train)
    _, disclosive = model.preliminary_check()
    assert not disclosive

    fname = "delete-me"
    metadata = model.run_attack(the_data, "worst_case", fname)
    files_made = ("delete-me.json", "log_roc.png")
    for fname in files_made:
        cleanup_file(fname)
    assert len(metadata) > 0  # something has been added


def test_run_attack_attribute():
    """calls the attribute  attack via safemodel"""
    the_data = get_nursery_dataset()
    assert the_data.__str__() == "nursery"

    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5)
    model.fit(the_data.x_train, the_data.y_train)
    _, disclosive = model.preliminary_check()
    assert not disclosive

    fname = "delete-me"
    metadata = model.run_attack(the_data, "attribute", fname)
    files_made = (
        "delete-me.json",
        "aia_example.json",
        "aia_example.pdf",
        "aia_report_cat_frac.png",
        "aia_report_cat_risk.png",
        "aia_report_quant_risk.png",
    )
    for fname in files_made:
        cleanup_file(fname)
    assert len(metadata) > 0  # something has been added


def test_attack_args():
    """tests the attack arguments class"""
    fname = "aia_example"
    attack_args = attribute_attack.AttributeAttackArgs(report_name=fname)
    attack_args.set_param("foo", "boo")
    assert attack_args.get_args()["foo"] == "boo"
    assert fname in attack_args.__str__()

    fname = "liraa"
    attack_args = likelihood_attack.LIRAAttackArgs(report_name=fname)
    attack_args.set_param("foo", "boo")
    assert attack_args.get_args()["foo"] == "boo"
    assert fname in attack_args.__str__()

    fname = "wca"
    attack_args = worst_case_attack.WorstCaseAttackArgs(report_name=fname)
    attack_args.set_param("foo", "boo")
    assert attack_args.get_args()["foo"] == "boo"
    assert fname in attack_args.__str__()


def test_run_attack_unknown():
    """calls an unknown attack via safemodel"""
    the_data = get_nursery_dataset()
    assert the_data.__str__() == "nursery"

    # build a model
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5)
    model.fit(the_data.x_train, the_data.y_train)

    fname = "delete-me"
    metadata = model.run_attack(the_data, "unknown", fname)
    files_made = ("delete-me.json",)
    for fname in files_made:
        cleanup_file(fname)
    assert metadata["outcome"] == "unrecognised attack type requested"



"""
Test the various models we have defined
"""

import unittest

import numpy as np
from sklearn import datasets
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

from aisdc.safemodel.classifiers import SafeSVC


def get_data():
    """Returns data for testing."""
    cancer = datasets.load_breast_cancer()
    x = np.asarray(cancer["data"], dtype=np.float64)
    y = np.asarray(cancer["target"], dtype=np.float64)

    return x, y


class TestDPSVC(unittest.TestCase):
    """
    Test the differentially private SVC
    """

    def test_run(self):
        """
        Test the model runs
        """
        dpsvc = SafeSVC()
        svc = SVC(kernel="rbf", gamma="scale", C=1.0, probability=True)
        x, y = get_data()

        dpsvc.fit(x, y)
        svc.fit(x, y)

        test_features = x
        dp_predictions = dpsvc.predict(test_features)
        sv_predictions = svc.predict(test_features)

        # Check that the two models have equal shape
        self.assertTupleEqual(dp_predictions.shape, sv_predictions.shape)

        dp_predprob = dpsvc.predict_proba(test_features)
        sv_predprob = svc.predict_proba(test_features)

        # Check that the two models have equal shape
        self.assertTupleEqual(dp_predprob.shape, sv_predprob.shape)

    def test_svc_recommended(self):
        """SafeSupportVectorClassifier using recommended values."""
        x, y = get_data()
        model = SafeSVC(gamma=1.0)
        model.fit(x, y)
        msg, disclosive = model.preliminary_check()
        correct_msg = "Model parameters are within recommended ranges.\n"
        assert msg == correct_msg
        assert disclosive is False

    def test_svc_khat(self):
        """SafeSupportVectorClassifier khat method."""
        x, y = get_data()
        model = SafeSVC(gamma=1.0)
        model.fit(x, y)
        y_matrix = np.ones((len(y), 30))
        assert len(y_matrix.shape) == 2
        _ = model.k_hat_svm(x, y_matrix)

    def test_svc_wrongdata(self):
        """SafeSupportVectorClassifier with wrong datatypes."""
        x, y = get_data()
        model = SafeSVC(gamma=1.0)
        # wrong y datatype
        with self.assertRaises(Exception) as context:
            model.fit(x, 1)
        self.assertTrue("DPSCV needs np.ndarray inputs" in str(context.exception))

        # wrong x datatype
        with self.assertRaises(Exception) as context:
            model.fit(1, 1)
        self.assertTrue("DPSCV needs np.ndarray inputs" in str(context.exception))

        # wrong label values
        yplus = y + 5
        errstr = "DP SVC can only handle binary classification"
        with self.assertRaises(Exception) as context:
            model.fit(x, yplus)
        self.assertTrue(errstr in str(context.exception))

    def test_svc_gamma_zero(self):
        """SafeSupportVectorClassifier still makes predictions if we provide daft params"""
        x, y = get_data()
        model = SafeSVC(gamma=0.0, eps=0.0)
        model.fit(x, y)
        predictions = model.predict(x)
        assert len(predictions) == len(x)

    def test_svc_gamma_auto(self):
        """SafeSupportVectorClassifier still makes predictions if we provide daft params"""
        x, y = get_data()
        model = SafeSVC(gamma="auto")
        model.fit(x, y)
        assert model.gamma == 1.0 / x.shape[1]

    def test_svc_setparams(self):
        """SafeSupportVectorClassifier using unchanged values."""
        x, y = get_data()
        model = SafeSVC(gamma=1.0)
        model.fit(x, y)

        newvals = {"gamma": 5, "eps": 0, "dhat": 999}
        model.set_params(**newvals)
        model.fit(x, y)
        assert model.eps == 0
        assert model.lambdaval == 0
        assert model.dhat == 999
        assert model.gamma == 5

        # unknown
        unrecognised = {"foo": 99}
        model.set_params(**unrecognised)
        # should log to file mewssage "Unsupported parameter: foo"

    def test_svc_unchanged(self):
        """SafeSupportVectorClassifier using unchanged values."""
        x, y = get_data()
        model = SafeSVC(gamma=1.0)
        model.fit(x, y)

        msg, disclosive = model.posthoc_check()
        correct_msg = ""
        assert msg == correct_msg
        assert disclosive is False

    def test_svc_nonstd_params_changed_postfit(self):
        """SafeSupportVectorClassifier with params changed after fit."""
        x, y = get_data()
        model = SafeSVC(gamma=1.0)
        model.fit(x, y)

        # change values
        model.svc = SVC().fit(x, y)  # make it disclosive
        model.platt_transform = LogisticRegression()
        msg, disclosive = model.posthoc_check()
        correct_msg = (
            "structure platt_transform has one difference.\n"
            "structure svc has several differences.\n"
        )

        assert msg == correct_msg
        assert disclosive is True



"""attack.py - base class for an attack object"""

import sklearn

from aisdc.attacks.dataset import Data


class Attack:
    """Base (abstract) class to represent an attack"""

    def attack(self, dataset: Data, target_model: sklearn.base.BaseEstimator) -> None:
        """Method to run an attack"""
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError



"""Jim Smith 2022
    Not currently implemented
    """

import pytest

from aisdc.safemodel.classifiers import safetf


def test_Safe_tf_DPModel_l2_and_noise():
    """Tests user is informed this is not implemented yet"""
    with pytest.raises(NotImplementedError):
        # with values for the l2 and noise params
        safetf.Safe_tf_DPModel(1.5, 2.0, True)



"""This module contains prototypes of privacy safe model wrappers."""

from __future__ import annotations

import copy
import datetime
import getpass
import json
import logging
import os
import pathlib
import pickle
from pickle import PicklingError
from typing import Any

import joblib

# import tensorflow as tf
from dictdiffer import diff

from aisdc.attacks import attribute_attack, dataset, report, worst_case_attack
from aisdc.attacks.likelihood_attack import (  # pylint: disable = import-error
    LIRAAttack,
    LIRAAttackArgs,
)

# pylint : disable=too-many-branches
from .reporting import get_reporting_string

logger = logging.getLogger(__file__)
logger.setLevel(logging.DEBUG)


def check_min(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Checks minimum value constraint.

    Parameters
    ----------

    key: string
         The dictionary key to examine.
    val: Any Type
         The expected value of the key.
    cur_val: Any Type
         The current value of the key.
    ..

    Returns
    -------

    msg: string
         A message string.
    disclosive: bool
         A boolean value indicating whether the model is potentially disclosive.

    Notes
    -----


    """
    if isinstance(cur_val, (int, float)):
        if cur_val < val:
            disclosive = True
            msg = get_reporting_string(
                name="less_than_min_value", key=key, cur_val=cur_val, val=val
            )
        else:
            disclosive = False
            msg = ""
        return msg, disclosive

    disclosive = True
    msg = get_reporting_string(
        name="different_than_recommended_type", key=key, cur_val=cur_val, val=val
    )
    return msg, disclosive


def check_max(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Checks maximum value constraint.

    Parameters
    ----------

    key: string
         The dictionary key to examine.
    val: Any Type
         The expected value of the key.
    cur_val: Any Type
         The current value of the key.

    Returns
    -------

    msg: string
         A message string.
    disclosive: bool
         A boolean value indicating whether the model is potentially disclosive.


    Notes
    -----


    """
    if isinstance(cur_val, (int, float)):
        if cur_val > val:
            disclosive = True
            msg = get_reporting_string(
                name="greater_than_max_value", key=key, cur_val=cur_val, val=val
            )
        else:
            disclosive = False
            msg = ""
        return msg, disclosive

    disclosive = True
    msg = get_reporting_string(
        name="different_than_recommended_type", key=key, cur_val=cur_val, val=val
    )
    return msg, disclosive


def check_equal(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Checks equality value constraint.



    Parameters
    ----------

    key: string
         The dictionary key to examine.
    val: Any Type
         The expected value of the key.
    cur_val: Any Type
         The current value of the key.

    Returns
    -------

    msg: string
         A message string.
    disclosive: bool
         A boolean value indicating whether the model is potentially disclosive.


    Notes
    -----


    """
    if cur_val != val:
        disclosive = True
        msg = get_reporting_string(
            name="different_than_fixed_value", key=key, cur_val=cur_val, val=val
        )
    else:
        disclosive = False
        msg = ""
    return msg, disclosive


def check_type(key: str, val: Any, cur_val: Any) -> tuple[str, bool]:
    """Checks the type of a value.

    Parameters
    ----------

    key: string
         The dictionary key to examine.
    val: Any Type
         The expected value of the key.
    cur_val: Any Type
         The current value of the key.

    Returns
    -------

    msg: string
         A message string.
    disclosive: bool
         A boolean value indicating whether the model is potentially disclosive.

    Notes
    -----


    """
    if type(cur_val).__name__ != val:
        disclosive = True
        msg = get_reporting_string(
            name="different_than_recommended_type", key=key, cur_val=cur_val, val=val
        )
    else:
        disclosive = False
        msg = ""
    return msg, disclosive


class SafeModel:  # pylint: disable = too-many-instance-attributes
    """Privacy protected model base class.

    Attributes
    ----------

    model_type: string
          A string describing the type of model. Default is "None".
    model:
          The Machine Learning Model.
    saved_model:
          A saved copy of the Machine Learning Model used for comparison.
    ignore_items: list
          A list of items to ignore when comparing the model with the
          saved_model.
    examine_separately_items: list
          A list of items to examine separately. These items are more
          complex datastructures that cannot be compared directly.
    filename: string
          A filename to save the model.
    researcher: string
          The researcher user-id used for logging



    Notes
    -----

    Examples
    --------
    >>> safeRFModel = SafeRandomForestClassifier()
    >>> safeRFModel.fit(X, y)
    >>> safeRFModel.save(name="safe.pkl")
    >>> safeRFModel.preliminary_check()
    >>> safeRFModel.request_release(filename="safe.pkl")
    WARNING: model parameters may present a disclosure risk:
    - parameter min_samples_leaf = 1 identified as less than the recommended min value of 5.
    Changed parameter min_samples_leaf = 5.

    Model parameters are within recommended ranges.




    """

    def __init__(self) -> None:
        """Super class constructor, gets researcher name."""
        self.model_type: str = "None"
        self.model = None
        self.saved_model = None
        self.model_load_file: str = "None"
        self.model_save_file: str = "None"
        self.ignore_items: list[str] = []
        self.examine_seperately_items: list[str] = []
        self.basemodel_paramnames = []
        self.filename: str = "None"
        self.researcher: str = "None"
        try:
            self.researcher = getpass.getuser()
        except (ImportError, KeyError, OSError):  # pragma: no cover
            self.researcher = "unknown"

    def get_params(self, deep=True):
        """gets dictionary of parameter values
        restricted to those expected by base classifier.
        """
        the_params = {}
        for key, val in self.__dict__.items():
            if key in self.basemodel_paramnames:
                the_params[key] = val
        if deep:
            pass  # not implemented yet
        return the_params

    def save(
        self, name: str = "undefined"
    ) -> None:  # pylint: disable=too-many-branches
        """Writes model to file in appropriate format.

        Note this is overloaded in SafeKerasClassifer
        to deal with tensorflow specifics.

        Parameters
        ----------

        name: string
             The name of the file to save

        Returns
        -------

        Notes
        -----

        No return value


        Optimizer is deliberately excluded.
        To prevent possible to restart training and thus
        possible back door into attacks.
        """

        self.model_save_file = name
        if self.model_save_file == "undefined":
            print("You must input a name with extension to save the model.")
        else:
            thename = self.model_save_file.split(".")
            # print(f'in save(), parsed filename is {thename}')
            if len(thename) == 1:
                print("file name must indicate type as a suffix")
            else:
                suffix = self.model_save_file.split(".")[-1]

                if (
                    suffix == "pkl" and self.model_type != "KerasModel"
                ):  # save to pickle
                    with open(self.model_save_file, "wb") as file:
                        try:
                            pickle.dump(self, file)
                        except (TypeError, AttributeError, PicklingError) as type_err:
                            print(
                                "saving a .pkl file is unsupported for model type:"
                                f"{self.model_type}."
                                f"Error message was {type_err}"
                            )

                elif (
                    suffix == "sav" and self.model_type != "KerasModel"
                ):  # save to joblib
                    try:
                        joblib.dump(self, self.model_save_file)
                    except (TypeError, AttributeError, PicklingError) as type_err:
                        print(
                            "saving as a .sav (joblib) file is not supported "
                            f"for models of type {self.model_type}."
                            f"Error message was {type_err}"
                        )
                #                  Overloaded in safekeras
                #                 elif suffix in ("h5", "tf") and self.model_type == "KerasModel":
                #                     try:
                #                         tf.keras.models.save_model(
                #                             self,
                #                             self.model_save_file,
                #                             include_optimizer=False,
                #                             # save_traces=False,
                #                             save_format=suffix,
                #                         )

                #                     except (ImportError, NotImplementedError) as exception_err:
                #                         print(
                #                             "saving as a {suffix} file gave this error message:"
                #                             f"{exception_err}"
                #                         )
                else:
                    print(
                        f"{suffix} file suffix currently not supported "
                        f"for models of type {self.model_type}.\n"
                    )

    ## Load functionality not needed
    # - provide directly by underlying pickle/joblib mechanisms
    # and safekeras provides its own to deal with tensorflow

    #     def load(self, name: str = "undefined") -> None:
    #         """reads model from file in appropriate format.
    #         Note that safekeras overloads this function.

    #         Optimizer is deliberately excluded in the save
    #         To prevent possible to restart training and thus
    #         possible back door into attacks.
    #         Thus optimizer cannot be loaded.
    #         """
    #         temp_file=None
    #         self.model_load_file = name
    #         if self.model_load_file == "undefined":
    #             print("You must input a file name with extension to load a model.")
    #         else:
    #             thename = self.model_save_file.split(".")
    #             suffix = self.model_save_file.split(".")[-1]

    #             if suffix == ".pkl":  # load from pickle
    #                 with open(self.model_load_file, "rb") as file:
    #                     temp_file = pickle.load(self, file)
    #             elif suffix == ".sav":  # load from joblib
    #                 temp_file = joblib.load(self, self.model_save_file)
    #             #safekeras overloads loads
    #             elif suffix in ("h5","tf")  and self.model_type != "KerasModel":
    #                 print("tensorflow objects saved as h5 or tf"
    #                       "can only be loaded into models of type SafeKerasClassifier"
    #                      )
    #             else:
    #                 print(f"loading from a {suffix} file is currently not supported")

    #         return temp_file

    def __get_constraints(self) -> dict:
        """Gets constraints relevant to the model type from the master read-only file."""
        rules: dict = {}
        rule_path = pathlib.Path(__file__).with_name("rules.json")
        with open(rule_path, encoding="utf-8") as json_file:
            parsed = json.load(json_file)
            rules = parsed[self.model_type]
        return rules["rules"]

    def __apply_constraints(
        self, operator: str, key: str, val: Any, cur_val: Any
    ) -> str:
        """Applies a safe rule for a given parameter."""
        if operator == "is_type":
            if (val == "int") and (type(cur_val).__name__ == "float"):
                self.__dict__[key] = int(self.__dict__[key])
                msg = get_reporting_string(name="change_param_type", key=key, val=val)
            elif (val == "float") and (type(cur_val).__name__ == "int"):
                self.__dict__[key] = float(self.__dict__[key])
                msg = get_reporting_string(name="change_param_type", key=key, val=val)
            else:
                msg = get_reporting_string(
                    name="not_implemented_for_change", key=key, cur_val=cur_val, val=val
                )
        else:
            setattr(self, key, val)
            msg = get_reporting_string(name="changed_param_equal", key=key, val=val)
        return msg

    def __check_model_param(
        self, rule: dict, apply_constraints: bool
    ) -> tuple[str, bool]:
        """Checks whether a current model parameter violates a safe rule.
        Optionally fixes violations."""
        disclosive: bool = False
        msg: str = ""
        operator: str = rule["operator"]
        key: str = rule["keyword"]
        val: Any = rule["value"]
        cur_val: Any = getattr(self, key)
        if operator == "min":
            msg, disclosive = check_min(key, val, cur_val)
        elif operator == "max":
            msg, disclosive = check_max(key, val, cur_val)
        elif operator == "equals":
            msg, disclosive = check_equal(key, val, cur_val)
        elif operator == "is_type":
            msg, disclosive = check_type(key, val, cur_val)
        else:
            msg = get_reporting_string(
                name="unknown_operator", key=key, val=val, cur_val=cur_val
            )
        if apply_constraints and disclosive:
            msg += self.__apply_constraints(operator, key, val, cur_val)
        return msg, disclosive

    def __check_model_param_and(
        self, rule: dict, apply_constraints: bool
    ) -> tuple[str, bool]:
        """Checks whether current model parameters violate a logical AND rule.
        Optionally fixes violations."""
        disclosive: bool = False
        msg: str = ""
        for arg in rule["subexpr"]:
            temp_msg, temp_disc = self.__check_model_param(arg, apply_constraints)
            msg += temp_msg
            if temp_disc:
                disclosive = True
        return msg, disclosive

    def __check_model_param_or(self, rule: dict) -> tuple[str, bool]:
        """Checks whether current model parameters violate a logical OR rule."""
        disclosive: bool = True
        msg: str = ""
        for arg in rule["subexpr"]:
            temp_msg, temp_disc = self.__check_model_param(arg, False)
            msg += temp_msg
            if not temp_disc:
                disclosive = False
        return msg, disclosive

    def preliminary_check(
        self, verbose: bool = True, apply_constraints: bool = False
    ) -> tuple[str, bool]:
        """Checks whether current model parameters violate the safe rules.
        Optionally fixes violations.


        Parameters
        ----------

        verbose: bool
             A boolean value to determine increased output level.

        apply_constraints: bool
             A boolean to determine whether identified constraints are
             to be upheld and applied.

        Returns
        -------

        msg: string
           A message string
        disclosive: bool
           A boolean value indicating whether the model is potentially
           disclosive.


        Notes
        -----


        """
        disclosive: bool = False
        msg: str = ""
        notok_start = get_reporting_string(name="warn_possible_disclosure_risk")
        ok_start = get_reporting_string(name="within_recommended_ranges")
        rules: dict = self.__get_constraints()
        for rule in rules:
            operator = rule["operator"]
            if operator == "and":
                temp_msg, temp_disc = self.__check_model_param_and(
                    rule, apply_constraints
                )
            elif operator == "or":
                temp_msg, temp_disc = self.__check_model_param_or(rule)
            else:
                temp_msg, temp_disc = self.__check_model_param(rule, apply_constraints)
            msg += temp_msg
            if temp_disc:
                disclosive = True

        if disclosive:
            msg = notok_start + msg
        else:
            msg = ok_start + msg

        if verbose:
            print("Preliminary checks: " + msg)
        return msg, disclosive

    def get_current_and_saved_models(self) -> tuple[dict, dict]:
        """Makes a copy of self.__dict__
        and splits it into dicts for the current and saved versions
        """
        current_model = {}

        attribute_names_as_list = copy.copy(list(self.__dict__.keys()))

        for key in attribute_names_as_list:
            if key not in self.ignore_items:
                # logger.debug(f'copying {key}')
                try:
                    value = self.__dict__[key]  # jim added
                    current_model[key] = copy.deepcopy(value)
                except (copy.Error, TypeError) as key_type:
                    logger.warning("%s cannot be copied", key)
                    logger.warning(
                        "...%s error; %s", str(type(key_type)), str(key_type)
                    )

        saved_model = current_model.pop("saved_model", "Absent")

        # return empty dict if necessary
        if (
            saved_model == "Absent"
            or saved_model is None
            or not isinstance(saved_model, dict)
        ):
            saved_model = {}
        else:
            # final check in case fit has been called twice
            _ = saved_model.pop("saved_model", "Absent")

        return current_model, saved_model

    def examine_seperate_items(
        self, curr_vals: dict, saved_vals: dict
    ) -> tuple[str, bool]:
        """comparison of more complex structures
        in the super class we just check these model-specific items exist
        in both current and saved copies"""
        msg = ""
        disclosive = False

        for item in self.examine_seperately_items:
            if curr_vals[item] == "Absent" and saved_vals[item] == "Absent":
                disclosive = True
                msg += get_reporting_string(name="both_item_removed", item=item)

            if curr_vals[item] == "Absent" and not saved_vals[item] == "Absent":
                msg += get_reporting_string(name="current_item_removed", item=item)
                disclosive = True

            if saved_vals[item] == "Absent" and not curr_vals[item] == "Absent":
                disclosive = True
                msg += get_reporting_string(name="saved_item_removed", item=item)

        if not disclosive:  # ok, so can call mode-specific extra checks
            msg, disclosive = self.additional_checks(curr_vals, saved_vals)

        return msg, disclosive

    def posthoc_check(self) -> tuple[str, bool]:  # pylint: disable=too-many-branches
        """Checks whether model has been interfered with since fit() was last run"""

        disclosive = False
        msg = ""

        current_model, saved_model = self.get_current_and_saved_models()
        if len(saved_model) == 0:
            msg = get_reporting_string(name="error_not_called_fit")
            msg += get_reporting_string(name="recommend_do_not_release")
            disclosive = True

        else:
            # remove things we don't care about
            for item in self.ignore_items:
                _ = current_model.pop(item, "Absent")
                _ = saved_model.pop(item, "Absent")

            # break out things that need to be handled/examined in more depth
            curr_separate = {}
            saved_separate = {}
            for item in self.examine_seperately_items:
                curr_separate[item] = current_model.pop(item, "Absent")
                saved_separate[item] = saved_model.pop(item, "Absent")

            # comparison on list of "simple" parameters
            match = list(diff(current_model, saved_model, expand=True))
            num_differences = len(match)
            if num_differences > 0:
                disclosive = True
                msg += get_reporting_string(
                    name="basic_params_differ", length=num_differences
                )
                for this_match in match:
                    if this_match[0] == "change":
                        msg += get_reporting_string(
                            name="param_changed_from_to",
                            key=this_match[1],
                            val=this_match[2][1],
                            cur_val=this_match[2][0],
                        )
                    else:
                        msg += f"{this_match}"

            # comparison on model-specific attributes
            extra_msg, extra_disclosive = self.examine_seperate_items(
                curr_separate, saved_separate
            )
            msg += extra_msg
            if extra_disclosive:
                disclosive = True

        return msg, disclosive

    def additional_checks(  # pylint: disable=too-many-branches
        self, curr_separate: dict, saved_separate: dict
    ) -> tuple[str, bool]:
        """Placeholder function for additional posthoc checks e.g. keras this
        version just checks that any lists have the same contents


        Parameters
        ----------

        curr_separate: python dictionary

        saved_separate: python dictionary


        Returns
        -------

        msg: string
        A message string
        disclosive: bool
        A boolean value to indicate whether the model is potentially disclosive.


        Notes
        -----

        posthoc checking makes sure that the two dicts have the same set of
        keys as defined in the list self.examine_separately

        """

        msg = ""
        disclosive = False
        for item in self.examine_seperately_items:
            if isinstance(curr_separate[item], list):
                if len(curr_separate[item]) != len(saved_separate[item]):
                    msg += (
                        f"Warning: different counts of values for parameter {item}.\n"
                    )
                    disclosive = True
                else:
                    for i in range(len(saved_separate[item])):
                        difference = list(
                            diff(curr_separate[item][i], saved_separate[item][i])
                        )
                        if len(difference) > 0:
                            msg += (
                                f"Warning: at least one non-matching value "
                                f"for parameter list {item}.\n"
                            )
                            disclosive = True
                            break

        return msg, disclosive

    def request_release(
        self, filename: str = "undefined", data_obj: dataset.Data = None
    ) -> None:  # pylint: disable=too-many-branches
        """Saves model to filename specified and creates a report for the TRE
        output checkers.

        Parameters
        ----------

        filename: string
        The filename used to save the model

        dataobj: object of type Data
        Contains train/test data and encoding dictionary needed to run attacks

        Returns
        -------


        Notes
        -----
         1. The dataset object is saved in a file called filebase_data.json
         (where filebase= filename without the extension)
         for reference/use by the TRE.
         Data should never be held or stored with the model.
         Clearly filebase_data.json mst never leave the TRE.
         2. If data_obj is not null, then worst case MIA and attribute inference
         attacks are called via run_attack.
         Outputs from the attacks will be stored in filebase_attack_res.json



        """
        if filename == "undefined":
            print("You must provide the name of the file you want to save your model")
            print("For security reasons, this will overwrite previous versions")
        else:
            self.save(filename)
            msg_prel, disclosive_prel = self.preliminary_check(verbose=False)
            msg_post, disclosive_post = self.posthoc_check()

            output: dict = {
                "researcher": self.researcher,
                "model_type": self.model_type,
                "model_save_file": self.model_save_file,
                "details": msg_prel,
            }
            if hasattr(self, "k_anonymity"):
                output["k_anonymity"] = f"{self.k_anonymity}"
            if not disclosive_prel and not disclosive_post:
                output[
                    "recommendation"
                ] = f"Run file {filename} through next step of checking procedure"
            else:
                output["recommendation"] = "Do not allow release"
                output["reason"] = msg_prel + msg_post

            ##Run attacks programmatically if possible
            if data_obj is not None:
                # make filenames and save a copy of the data
                with open(
                    os.path.splitext(filename)[0] + "_data.pickle", "wb"
                ) as handle:
                    pickle.dump(data_obj, handle, protocol=pickle.HIGHEST_PROTOCOL)

                for attack_name in ["worst_case", "lira", "attribute"]:
                    output[f"{attack_name}_results"] = self.run_attack(
                        data_obj,
                        attack_name,
                        f"{os.path.splitext(filename)[0]}_{attack_name}_res",
                    )

            now = datetime.datetime.now()
            output["timestamp"] = str(now.strftime("%Y-%m-%d %H:%M:%S"))

            json_str = json.dumps(output, indent=4, cls=report.NumpyArrayEncoder)
            outputfilename = self.researcher + "_checkfile.json"
            with open(outputfilename, "a", encoding="utf-8") as file:
                file.write(json_str)

    def run_attack(
        self,
        data_obj: dataset.Data = None,
        attack_name: str = "worst_case",
        filename: str = "undefined",
    ) -> dict:
        """Runs a specified attack on the trained model and saves a report to file

        Parameters
        ----------
        data_obj: Data
        the dataset in the form of a Data object

        attack_name: string

        filebasename: string
        Report will be saved to filebasename.json


        Returns
        -------
        dict of meta data results

        Notes
        -----
        Currently implement attack types are:
        Likelihood Ratio: lira
        Worst_Case Membership inference: worst_case
        Single Attribute Inference: attributes
        """
        if attack_name == "worst_case":
            attack_args = worst_case_attack.WorstCaseAttackArgs(
                n_reps=10,
                # number of baseline (dummy) experiments to do
                n_dummy_reps=1,
                # Threshold to determine significance of things
                p_thresh=0.05,
                # Filename arguments needed by the code, meaningless if run programmatically
                in_sample_filename=None,
                out_sample_filename=None,
                # Proportion of data to use as a test set for the attack model;
                test_prop=0.5,
                # Report name is None - don't make json or pdf files
                report_name=None,
            )
            attack_obj = worst_case_attack.WorstCaseAttack(attack_args)
            attack_obj.attack(dataset=data_obj, target_model=self)
            output = attack_obj.make_report()
            metadata = output["metadata"]

        elif attack_name == "lira":
            args = LIRAAttackArgs(
                n_shadow_models=100, report_name="lira_example_report"
            )
            attack_obj = LIRAAttack(args)
            attack_obj.attack(data_obj, self)
            output = attack_obj.make_report()  # also makes .pdf and .json files
            metadata = output["metadata"]

        elif attack_name == "attribute":
            attack_args = attribute_attack.AttributeAttackArgs(
                report_name="aia_example"
            )
            attack_obj = attribute_attack.AttributeAttack(attack_args)
            attack_obj.attack(data_obj, self)
            output = attack_obj.make_report()
            metadata = output["metadata"]

        else:
            metadata = {}
            metadata["outcome"] = "unrecognised attack type requested"

        print(f"attack {attack_name}, metadata {metadata}")

        try:
            with open(f"{filename}.json", "w", encoding="utf-8") as fp:
                json.dump(metadata, fp, cls=report.NumpyArrayEncoder)
        except TypeError:  # pragma: no cover
            # not covered in tests as all atttacks prodice simple json so far
            print(f"couldn't serialise metadata {metadata} for attack {attack_name}")

        return metadata

    def __str__(self) -> str:  # pragma: no cover
        """Returns string with model description.
        No point writing a test, especially as it depends on username
        """
        return self.model_type + " with parameters: " + str(self.__dict__)
